import bottleneck as bn
import numpy as np

from hotspot3.io.logging import WithLogger
from hotspot3.helpers.utils import wrap_masked, correct_offset


class BottleneckWrapper(WithLogger):

    @wrap_masked
    @correct_offset
    def centered_running_nansum(self, array: np.ndarray, window: int) -> np.ndarray:
        min_count = self.get_min_count(window)
        return bn.move_sum(array, window, min_count=min_count).astype(np.float32)

    @wrap_masked
    @correct_offset
    def centered_running_nanvar(self, array, window) -> np.ndarray:
        min_count = self.get_min_count(window)
        return bn.move_var(array, window, ddof=1, min_count=min_count).astype(np.float32)

    @wrap_masked
    @correct_offset
    def centered_running_nanmean(self, array, window) -> np.ndarray:
        min_count = self.get_min_count(window)
        return bn.move_mean(array, window, min_count=min_count).astype(np.float32)

    @wrap_masked
    @correct_offset
    def centered_running_nanmax(self, array, window)  -> np.ndarray:
        return bn.move_max(array, window, min_count=1).astype(np.float32)

    def get_min_count(self, window):
        return max(1, round(window * self.config.min_mappable_bases_proportion))
    
    @wrap_masked
    def get_max_count_with_flanks(self, array: np.ndarray, flank_length: int=None):
        if flank_length is None:
            flank_length = self.config.exclude_peak_flank_fit
        return self.centered_running_nanmax(array, flank_length * 2 + 1)
    
    @wrap_masked
    def get_signal_mask_for_tr(self, array: np.ndarray, tr: float, flank_length: int=None):
        max_count = self.get_max_count_with_flanks(array, flank_length=flank_length)
        return max_count >= tr


import numpy as np
import itertools

from hotspot3.io.logging import WithLogger
from hotspot3.helpers.models import WindowedFitResults
from hotspot3.scoring import logpval_for_dtype, fix_inf_pvals


class PvalueEstimator(WithLogger):

    def estimate_pvalues(self, agg_cutcounts: np.ndarray, fit_results: WindowedFitResults) -> np.ndarray:
        result = np.full_like(agg_cutcounts, np.nan, dtype=np.float16)
        r = fit_results.r
        p = fit_results.p
        mask = fit_results.enough_bg_mask
        data, invalid = self.negbin_neglog10pvalue(agg_cutcounts[mask], r[mask], p[mask])
        result[mask] = data
        if invalid is not None:
            ids = np.arange(result.shape[0])[mask][invalid]
            self.logger.critical(f"{self.name}: {len(ids)} p-values are NaN for betainc method, {ids}. Parameters: r={r[ids]}, p={p[ids]}, count={agg_cutcounts[ids]}")
            raise ValueError(f"{self.name}: {len(ids)} p-values are NaN for betainc method, {ids}")
        return result

    def negbin_neglog10pvalue(self, x: np.ndarray, r: np.ndarray, p: np.ndarray) -> np.ndarray:
        result = logpval_for_dtype(x, r, p, dtype=np.float32, calc_type="betainc").astype(np.float16)
        low_precision = np.isinf(result)
        invalid_pvals = np.isnan(result)
        if np.any(invalid_pvals):
            return result, invalid_pvals
        for precision, method in itertools.product(
            (np.float32, np.float64),
            ("betainc", "hyp2f1", "nbinom")
        ):
            if precision == np.float32 and method == "betainc":
                continue
            if np.any(low_precision):
                new_pvals = logpval_for_dtype(
                    x[low_precision],
                    r[low_precision],
                    p[low_precision],
                    dtype=precision,
                    calc_type=method
                )
                corrected_infs = np.isfinite(new_pvals)
                result[low_precision] = np.where(corrected_infs, new_pvals, result[low_precision])
                low_precision[low_precision] = ~corrected_infs

            else:
                break
        result /= -np.log(10).astype(result.dtype)
        return result, None

    def fix_inf_pvals(self, pvals):
        return fix_inf_pvals(pvals)


from hotspot3.scoring.fdr import MultiSampleFDRCorrection
from hotspot3.config import ProcessorConfig
from hotspot3.io.readers import GenomeReader

import pandas as pd
import logging
import argparse

# TODO: Finish implementing this function
def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('save_path', type=str)
    parser.add_argument('--mapping_df', type=str, default=None)

    parser.add_argument('--fdr_cutoff', type=float, default=0.05)
    parser.add_argument('--chrom_sizes', type=str, default=None)
    parser.add_argument('--cpus', type=int, default=10)

    return parser.parse_args()

def main():
    args = parse_args()
    config = ProcessorConfig(cpus=args.cpus, logger_level=logging.DEBUG)
    reader = GenomeReader(config=config)
    chrom_sizes = reader.read_chrom_sizes(args.chrom_sizes)
    mapping_df = pd.read_table(args.mapping_df).set_index('id')['pvals_parquet']
    reader.copy_with_params(
        MultiSampleFDRCorrection,
        name=mapping_df.index.tolist(),
        chrom_sizes=chrom_sizes
    ).fdr_correct_pvals(
        paths=mapping_df.to_dict(),
        fdr_cutoff=args.fdr_cutoff,
        save_path=args.save_path
    )


import pandas as pd
import numpy as np
import functools
import numpy.ma as ma

from hotspot3.helpers.models import NotEnoughDataForContig


def ensure_contig_exists(func):
    """
    Decorator for functions that require a contig to be present in the input data.

    Returns None if the contig is not present.
    Otherwise, returns the result of the function.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except NotEnoughDataForContig:
            return None
    return wrapper


def is_iterable(obj):
    if isinstance(obj, pd.DataFrame) or isinstance(obj, str):
        return False
    try:
        iter(obj)
        return True
    except TypeError:
        return False


def wrap_masked(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        masked_arrays = [arg for arg in args if isinstance(arg, ma.MaskedArray)]
        masked_arrays += [value for value in kwargs.values() if isinstance(value, ma.MaskedArray)]

        if len(masked_arrays) > 0: # introduces overhead, maybe ignore this check
            assert all(masked_arrays[0].shape == ma.shape for ma in masked_arrays), 'All masked arrays should have the same shape'
            assert all(np.all(masked_arrays[0].mask == ma.mask) for ma in masked_arrays), 'All masked arrays should have the same mask'
            mask = masked_arrays[0].mask
        else:
            return func(*args, **kwargs)
        args = [compress_masked_arg(arg) for arg in args]
        kwargs = {key: compress_masked_arg(value) for key, value in kwargs.items()}
        result = func(*args, **kwargs)

        if isinstance(result, np.ndarray):
            return wrap_if_1d(mask, result)
        elif isinstance(result, tuple):
            return tuple(wrap_if_1d(mask, r) for r in result)
        else:
            return result
    return wrapper

def wrap_if_1d(mask: np.ndarray, result: np.ndarray):
    if result.ndim == 0:
        return result
    return ma.masked_where(mask, result)

def compress_masked_arg(arg):
    if isinstance(arg, ma.MaskedArray):
        return arg.filled(np.nan)
    else:
        return arg


def correct_offset(func):
    """
    Correct offset of a trailing running window to make it centered.
    """
    @functools.wraps(func)
    def wrapper(self, array, window, *args, **kwargs):
        assert window % 2 == 1, "Window size should be odd"
        if len(array) < window:
            window = len(array)
        array = np.asarray(array, dtype=np.float32)
        offset = window // 2
        result = func(
            self,
            np.pad(array, (0, offset), mode='constant', constant_values=np.nan),
            window,
            *args,
            **kwargs
        )
        return result[offset:]
    return wrapper


@wrap_masked
def interpolate_nan(array):
    subsampled_indices = np.where(~np.isnan(array))[0]
    subsampled_arr = array[subsampled_indices]
    try:
        return np.interp(
            np.arange(array.shape[0], dtype=np.uint32),
            subsampled_indices,
            subsampled_arr,
            left=None,
            right=None,
        )
    except ValueError:
        return np.full(array.shape, np.nan, dtype=np.float32)



import numpy as np
import scipy.stats as st
import gc
from scipy.special import logsumexp, betainc, hyp2f1, betaln
from typing import Union

from hotspot3.helpers.models import WindowedFitResults, FitResults


# Calculate p-values and FDR
def logpval_for_dtype(x: np.ndarray, r: np.ndarray, p: np.ndarray, dtype=None, calc_type="betainc") -> np.ndarray:
    """
    Implementation of log(pval) with high precision
    """
    mask = x > 0
    x = np.round(np.asarray(x, dtype=dtype))
    r = np.broadcast_to(np.asarray(r, dtype=dtype), x.shape)[mask]
    p = np.broadcast_to(np.asarray(p, dtype=dtype), x.shape)[mask]
    x = x[mask]
    
    result = np.zeros(mask.shape, dtype=dtype)
    if calc_type == 'nbinom':
        p_vals = st.nbinom.logsf(x - 1, r, 1 - p)
    elif calc_type == 'betainc':
        p_vals = logpval_for_dtype_betainc(x, r, p)
    elif calc_type == 'hyp2f1':
        p_vals = logpval_for_dtype_hyp2f(x, r, p)
    else:
        raise ValueError(f"Unknown p-value calculation type: {calc_type}")
    result[mask] = p_vals
    return result


def logpval_for_dtype_betainc(x: np.ndarray, r: np.ndarray, p: np.ndarray) -> np.ndarray:
    with np.errstate(divide='ignore', invalid='ignore'):
        return np.log(betainc(x, r, p, dtype=r.dtype))


def logpval_for_dtype_hyp2f(x: np.ndarray, r: np.ndarray, p: np.ndarray) -> np.ndarray:
    with np.errstate(divide='ignore', invalid='ignore'):
        return (
            x * np.log(p) 
            + r * np.log(1 - p) 
            + np.log(hyp2f1(x + r, 1, x + 1, p, dtype=r.dtype))
            - np.log(x)
            - betaln(x, r, dtype=r.dtype)
        )


def logfdr_from_logpvals(log_pvals, *, method='bh', dtype=np.float32, m=None):
    """
    Reimplementation of scipy.stats.false_discovery_control to work with neglog-transformed p-values.
    Accepts log-transformed p-values and returns log-transformed FDR values.
    NOTE: Not log10-transformed and not negated!

    Parameters:
        - log_pvals: 1D array of log-transformed p-values. Most of them will be negative.
        - fdr_method: Method to use for FDR calculation. 'bh' = Benjamini-Hochberg, 'by' = Benjamini-Yekutieli.
    
    Returns:
        - log_fdr: 1D array of log-transformed FDR values. 
    """
    assert method in ['bh', 'by'], "Only 'bh' and 'by' methods are supported."
    log_pvals = np.asarray(log_pvals, dtype=dtype) # no further input checking yet
    if log_pvals.ndim != 1:
        raise NotImplementedError("Only 1D arrays are supported.")

    if m is None:
        m = log_pvals.shape[0]
    order = np.argsort(log_pvals) # can save argsort as uint32
    
    log_pvals = log_pvals[order]

    log_i = np.log(np.arange(1, log_pvals.shape[0] + 1, dtype=dtype))
    log_pvals += np.log(m).astype(dtype) - log_i  # p_adj = p * m / i => log10(p_adj) = log10(p) + log10(m) - log10(i)
    if method == 'by':
        log_pvals += logsumexp(-log_i)
    del log_i
    gc.collect()
    np.minimum.accumulate(log_pvals[::-1], out=log_pvals[::-1])
    log_pvals_copy = log_pvals.copy() # since array is float32, copying is more memory effient than argsort int64
    log_pvals[order] = log_pvals_copy
    return np.clip(log_pvals, a_min=None, a_max=0)


def calc_g_sq(obs, exp):
    valid = (exp != 0) & (obs != 0)
    with np.errstate(over='ignore'):
        ratio = np.divide(obs, exp, out=np.zeros_like(obs), where=valid)
    ratio = np.where(valid, ratio, 1)
    return obs * np.log(ratio) * 2


def calc_chisq(obs, exp):
    with np.errstate(over='ignore'):
        return np.where((exp != 0) & (obs != 0), (obs - exp) ** 2 / exp, 0)


def calc_rmsea(obs, exp, N, df, min_df=1, stat='G_sq', where=None):
    assert stat in ('G_sq', 'chi_sq'), "Only G_sq and chi_sq statistics are supported"
    if stat == 'G_sq':
        G_sq = calc_g_sq(obs, exp)
    else:
        G_sq = calc_chisq(obs, exp)

    G_sq = np.sum(G_sq, axis=0, where=where)
    G_sq = np.divide(G_sq, df, out=np.zeros_like(G_sq), where=df>=min_df)
    rmsea = np.sqrt(np.maximum(G_sq - 1, 0) / (N - 1))
    rmsea = np.where(df >= min_df, rmsea, np.inf)
    return rmsea


def check_valid_nb_params(fit: Union[WindowedFitResults, FitResults]):
    return (fit.r > 0.) & (fit.p > 0.) & (fit.p < 1.) 


def upper_bg_quantile(r, p, quantile=0.005):
    r = np.asarray(r, dtype=np.float32)
    p = np.asarray(p, dtype=np.float32)
    return st.nbinom(r, 1 - p).isf(quantile)


def mean_from_r_p(r, p):
    return r * p / (1 - p)


def weighted_median(data, weights):
    data = np.asarray(data)
    weights = np.asarray(weights)
    sorted_indices = np.argsort(data)
    sorted_data = data[sorted_indices]
    sorted_weights = weights[sorted_indices]
    cum_weights = np.cumsum(sorted_weights)
    median_index = np.searchsorted(cum_weights, 0.5 * cum_weights[-1])
    return sorted_data[median_index]


def weighted_std(x, weights, mean):
    weights = np.asarray(weights)
    x = np.asarray(x)
    variance = np.sum(weights * np.square(x - mean)) / np.sum(weights)
    return np.sqrt(variance)

def threhold_from_bg_tag_proportion(signal_at_segment, min_bg_tag_proportion: float) -> float:
    uq, cts = np.unique(signal_at_segment, return_counts=True)
    total = uq * cts
    cdf = np.cumsum(total) / np.sum(total)
    valid_cts = uq[cdf >= min_bg_tag_proportion]
    if valid_cts.size < 2: # Only at largest count min_bg_tag_proportion < cdf (cdf = 1)
        # Since background is defined as less (not leq) => tr - use inf
        return np.inf
    return valid_cts[1]

def roundup_float(number, decimals):
    return np.ceil(number * (10 ** decimals)) / (10 ** decimals)



import pandas as pd
import numpy as np
from sklearn.linear_model import RANSACRegressor
from scipy.special import logit, expit

from hotspot3.io.logging import WithLogger
from hotspot3.helpers.models import SPOTEstimationResults, SPOTEstimationData
from hotspot3.helpers.stats import upper_bg_quantile, weighted_median, weighted_std
from hotspot3.helpers.format_converters import get_spot_score_fit_data
from hotspot3.helpers.stats import roundup_float


class SignalToNoiseFit(WithLogger):

    def fit(self, fit_data: pd.DataFrame):
        """
        Fit the signal to the noise using a linear regression.
        """
        fit_data['background'] = upper_bg_quantile(
            fit_data['bg_r'],
            fit_data['bg_p']
        )
        valid_segment_fit = fit_data.eval('fit_type == "segment" & success_fit')

        spot_data = get_spot_score_fit_data(fit_data[valid_segment_fit])
        fit_data.loc[valid_segment_fit, 'segment_SPOT'] = spot_data.segment_spot_scores

        spot_results = self.fit_spot_length_regression(spot_data)

        # One value per segment
        fit_data.loc[valid_segment_fit, 'outlier_distance'] = spot_results.outlier_distance
        fit_data.loc[valid_segment_fit, 'is_inlier'] = spot_results.inliers_mask
        fit_data.loc[valid_segment_fit, 'min_bg_tag_proportion'] = spot_results.min_bg_tags_fraction
        
        # One value for the dataset
        fit_data['slope'] = spot_results.slope
        fit_data['intercept'] = spot_results.intercept
        fit_data['SPOT'] = spot_results.spot_score
        fit_data['SPOT_std'] = spot_results.spot_score_std

        return fit_data, spot_results
    
    def calc_dataset_spot_score(self, spot_data: SPOTEstimationData):
        spot_scores = spot_data.segment_spot_scores[spot_data.valid_scores]
        total_bases = spot_data.total_bases[spot_data.valid_scores]
        if len(spot_scores) == 0:
            self.logger.warning("No segments with valid SPOT score found for the dataset.")
            return np.nan
        spot_score = weighted_median(spot_scores, total_bases)
        return spot_score
    
    def calc_spot_score_error(self, resid, spot_score, total_bases):
        detrended_segment_spot = expit(resid + logit(spot_score))
        return weighted_std(detrended_segment_spot, total_bases, spot_score)
    
    def fit_spot_length_regression(self, spot_data: SPOTEstimationData):
        model = RANSACRegressor(random_state=42, min_samples=0.2)
        spot_score = self.calc_dataset_spot_score(spot_data)

        where = spot_data.valid_scores
        inliers = np.zeros_like(where, dtype=bool)

        if np.isnan(spot_score):
            return SPOTEstimationResults(
                spot_score=np.nan,
                spot_score_std=np.nan,
                inliers_mask=inliers,
                slope=np.nan,
                intercept=np.nan,
                outlier_distance=np.nan,
                min_bg_tags_fraction=1
            )

        X = np.log(spot_data.total_bases)[:, None]
        y = logit(spot_data.segment_spot_scores)


        model.fit(X[where], y[where])
    
        inliers[where] = model.inlier_mask_
        slope, intercept = model.estimator_.coef_[0], model.estimator_.intercept_
        
        y_pred = model.predict(X)
        resid = y - y_pred
        min_bg_tags_fraction = roundup_float(
            expit(
                -y_pred - np.log(self.config.outlier_segment_threshold)
            ), 
            5
        ) # Ceil to 5th decimal place to avoid precision issues
        outlier_dist = np.exp(resid)

        spot_score_std = self.calc_spot_score_error(resid, spot_score, spot_data.total_bases)

        self.logger.info(f"Signal to noise fit: SPOT={spot_score:.2f}±{spot_score_std:.2f}")

        return SPOTEstimationResults(
            spot_score=spot_score,
            spot_score_std=spot_score_std,
            inliers_mask=inliers,
            slope=slope,
            intercept=intercept,
            outlier_distance=outlier_dist,
            min_bg_tags_fraction=min_bg_tags_fraction
        )

    def find_outliers(self, fit_data: pd.DataFrame) -> np.ndarray:
        return fit_data.eval(f'outlier_distance >= {self.config.outlier_segment_threshold} & fit_type == "segment"').values



import subprocess
import importlib.resources as pkg_resources
from collections import defaultdict
import pysam
import pandas as pd
from io import StringIO

from hotspot3.io.logging import WithLogger
from hotspot3.helpers.models import NotEnoughDataForContig


def run_bam2bed(*args, reference_fasta=None):
    fasta_args = fasta_as_arg(reference_fasta)
    with pkg_resources.path('hotspot3.scripts', 'extract_cutcounts.sh') as script:
        cmd = f'set -e -o pipefail; bash {script} ' + " ".join(fasta_args) + " " + " ".join(args)
        try:
            result = subprocess.run(
                cmd,
                check=True,
                text=True,
                capture_output=True,
                shell=True,
            )
        except subprocess.CalledProcessError as e:
             raise RuntimeError(
                    f"extract_cutcounts.sh failed with exit code {e.returncode}.\n"
                    f"Command: {cmd}\n"
                    f"stderr:\n{e.stderr}")
    return result

def fasta_as_arg(reference_fasta):
    if reference_fasta is not None:
        return ('-T', reference_fasta)
    else:
        return ()

class BamFileCutsExtractor(WithLogger):
    def extract_all_chroms(self, bam_path, tabix_bed_path, chromosomes, reference_fasta=None):
        """
        Run bam2bed conversion script.
        Very fast but can't be parallelized.
        """
        run_bam2bed(bam_path, *chromosomes, '|', 'bgzip', '>', tabix_bed_path, reference_fasta=reference_fasta)
        pysam.tabix_index(tabix_bed_path, preset='bed', force=True)


    def extract_chromosome_to_df(self, bam_path, chromosome: str, reference_fasta=None) -> pd.DataFrame:
        """
        Run bam2bed for a single chromosome. Returns a pandas DataFrame.
        """
        result = run_bam2bed(bam_path, chromosome, reference_fasta=reference_fasta)
        df = pd.read_table(StringIO(result.stdout))
        return df.drop(columns=['#chr'])

    def extract_reads_pysam(self, bam_path, chromosome) -> pd.DataFrame:
        """
        Extract reads with pysam for a single chromosome.
        Slower than bam2bed conversion but can be heavily parallelized.
        """
        bed_counts = defaultdict(int)
        try:
            with pysam.AlignmentFile(bam_path) as bamfile:
                for read in bamfile.fetch(chromosome):
                    if read.is_unmapped or read.is_secondary or read.is_supplementary:
                        continue

                    if read.is_reverse:
                        cut_start = read.reference_end
                    else:
                        cut_start = read.reference_start
                    bed_counts[cut_start] += 1
        except ValueError:
            raise NotEnoughDataForContig

        bed_df = pd.DataFrame(
            [(start, count) for start, count in bed_counts.items()],
            columns=["start", "count"]
        ).sort_values("start")
        bed_df["end"] = bed_df["start"] + 1
        bed_df["chrom"] = chromosome

        return bed_df[['chrom', 'start', 'end', 'count']]


import argparse
import logging
import sys


epilog = """
Examples:
hotspot3 sample123 --bam sample123.bam --chrom_sizes hg38.chrom.sizes --fdrs 0.01 0.05
hotspot3 sample123 --chrom_sizes hg38.chrom.sizes --cutcounts sample123.cutcounts.bed.gz --signal_parquet sample123.signal.parquet --fdrs_parquet sample123.fdrs.parquet --fdrs 0.001

See full documentation at: https://github.com/vierstralab/hotspot3
"""

def parse_arguments(extra_desc: str = "") -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description="Call peaks using hotspot3 with local background estimation and MODWT smoothing.",
        epilog=epilog + extra_desc
    )
    
    # common arguments
    parser.add_argument("id", type=str, help="Unique identifier of the sample")
    parser.add_argument("--bam", help="Path to input bam/cram file", default=None)

    parser.add_argument(
        "--chrom_sizes", 
        help="Path to chromosome sizes file. If none assumed to be hg38 sizes",
        default=None
    )
    parser.add_argument(
        "--fdrs", 
        help="List of FDR thresholds, space separated",
        type=float, 
        nargs='+',
        default=[0.05]
    )
    parser.add_argument(
        "--reference",
        help="Path to reference fasta file. Required to work with cram files with missing fasta", default=None
    )
    parser.add_argument(
        "--cpus",
        type=int,
        help="Number of CPUs to use. Doesn't use more than # of chromosomes.",
        default=1
    )
    parser.add_argument(
        "--outdir",
        help="Path to output directory",
        default="."
    )
    parser.add_argument(
        "--debug",
        help="Adds additional prints",
        action='store_true',
        default=False
    )
    parser.add_argument(
        "--tempdir",
        help="Path to temporary directory. Defaults to system temp directory",
        default=None
    )

    parser.add_argument(
        "--mappable_bases", 
        help=" (Optional) Path to tabix indexed mappable bases file. Ignore unmappable bases for statistical model.",
        default=None
    )
    parser.add_argument(
        "--window",
        help="Window size for smoothing cutcounts",
        type=int,
        default=151
    )
    parser.add_argument(
        "--background_window",
        help="Background window size",
        type=int,
        default=50001
    )
    parser.add_argument(
        "--signal_quantile",
        help="Max proportion of background expected in the data. Change if you know what you are doing.",
        type=float,
        default=0.995
    )
    
    # Arguments to skip previous steps if provided

    parser.add_argument(
        "--cutcounts",
        help="Path to pre-calculated cutcounts tabix file. Skip extracting cutcounts from bam file",
        default=None
    )
    parser.add_argument(
        "--signal_parquet",
        help="Path to pre-calculated partitioned parquet file(s) with per-bp smoothed signal. Skips modwt signal smoothing",
        default=None
    )

    parser.add_argument(
        "--pvals_parquet",
        help="Path to pre-calculated partitioned parquet file(s) with per-bp p-values. Skips p-value calculation",
        default=None
    )

    parser.add_argument(
        "--fdrs_parquet",
        help="Path to pre-calculated fdrs. Can correct for several samples using multiple_samples_fdr.py",
        default=None
    )

    parser.add_argument(
        "--chromosomes",
        help="List of chromosomes to process. Useful for debug",
        nargs='+',
        default=None
    )

    # Optional - save density
    parser.add_argument(
        "--save_density",
        action='store_true',
        help="Save normalized density of cutcounts",
        default=False
    )

    args = parser.parse_args()
    return args


def main() -> None:
    args = parse_arguments()

    # Imports here to make help message appear faster
    from hotspot3.io.logging import setup_logger
    logger_level = logging.DEBUG if args.debug else logging.INFO
    root_logger = setup_logger(level=logger_level)
    root_logger.info(f"Executing command: {' '.join(sys.argv)}")


    from hotspot3.config import ProcessorConfig
    from hotspot3.from_configs import GenomeProcessor, run_from_configs
    from hotspot3.io.paths import Hotspot3Paths

    config = ProcessorConfig(
        window=args.window,
        bg_window=args.background_window,
        max_background_prop=args.signal_quantile,
        save_debug=args.debug,
        cpus=args.cpus,
        logger_level=logger_level,
        tmp_dir=args.tempdir,
    )
    genome_processor = GenomeProcessor(
        sample_id=args.id,
        chrom_sizes_file=args.chrom_sizes,
        mappable_bases_file=args.mappable_bases,
        chromosomes=args.chromosomes,
        reference_fasta=args.reference,
        config=config,
    )

    paths = Hotspot3Paths(
        outdir=args.outdir,
        sample_id=args.id,
        save_density=args.save_density,
        bam=args.bam,
        cutcounts=args.cutcounts,
        smoothed_signal=args.signal_parquet,
        pvals=args.pvals_parquet,
        fdrs=args.fdrs_parquet,
    )
    run_from_configs(genome_processor, paths, args.fdrs)
    
    root_logger.info('Program finished')


if __name__ == "__main__":
    main()



import os
import shutil

from hotspot3.processors import GenomeProcessor
from hotspot3.io.paths import Hotspot3Paths


def run_from_configs(
        genome_processor: GenomeProcessor, 
        paths: Hotspot3Paths, 
        fdrs,
    ):
    step_names = paths.find_missing_steps()
    genome_processor.logger.info(f"Running: {', '.join(paths.get_display_names(step_names))}")
    if 'cutcounts' in step_names:
        genome_processor.extract_cutcounts_from_bam(paths.bam, paths.cutcounts)
    
    if 'total_cutcounts' in step_names:
        genome_processor.get_total_cutcounts(paths.cutcounts, paths.total_cutcounts)
    
    if 'smoothed_signal' in step_names:
        genome_processor.smooth_signal_modwt(
            paths.cutcounts,
            save_path=paths.smoothed_signal,
            total_cutcounts_path=paths.total_cutcounts
        )
    if 'fit_params' in step_names:
        genome_processor.fit_background_model(
            paths.cutcounts,
            total_cutcounts_path=paths.total_cutcounts,
            save_path=paths.fit_params,
            per_region_stats_path=paths.per_region_stats,
            per_region_stats_path_bw=paths.per_region_background,
        )

        genome_processor.extract_fit_thresholds_to_bw(
            paths.fit_params,
            paths.total_cutcounts,
            paths.thresholds,
        )

        genome_processor.extract_bg_quantile_to_bw(
            paths.fit_params,
            paths.total_cutcounts,
            paths.background
        )
    
    if 'pvals' in step_names:
        genome_processor.calc_pvals(
            paths.cutcounts,
            paths.fit_params,
            paths.pvals
        )
    
    if 'fdrs' in step_names:
        genome_processor.calc_fdr(
            paths.pvals,
            paths.fdrs,
            max(fdrs)
        )
    
    if 'normalized_density' in step_names:
        genome_processor.extract_normalized_density(
            paths.smoothed_signal,
            paths.normalized_density
        )
    
    if 'peak_calling' in step_names:
        genome_processor.logger.info(f'Calling peaks and hotspots at FDRs: {fdrs}') 
        for fdr in fdrs:
            fdr_dir = paths.fdrs_dir(fdr)
            if os.path.exists(fdr_dir):
                shutil.rmtree(fdr_dir)
            os.makedirs(fdr_dir, exist_ok=True)

            genome_processor.logger.debug(f'Calling hotspots at FDR={fdr}')
            genome_processor.call_hotspots(
                paths.fdrs,
                sample_id=genome_processor.sample_id,
                save_path=paths.hotspots(fdr),
                save_path_bb=paths.hotspots_bb(fdr),
                fdr_tr=fdr
            )

            genome_processor.logger.debug(f'Calling variable width peaks at FDR={fdr}')
            genome_processor.call_variable_width_peaks(
                paths.smoothed_signal,
                paths.fdrs,
                paths.total_cutcounts,
                sample_id=genome_processor.sample_id,
                save_path=paths.peaks(fdr),
                save_path_bb=paths.peaks_bb(fdr),
                fdr_tr=fdr
            )



import numpy as np
import numpy.ma as ma
import pandas as pd
from typing import List

from genome_tools.data.extractors import TabixExtractor, ChromParquetExtractor
from genome_tools import GenomicInterval

from hotspot3.helpers.models import NotEnoughDataForContig, WindowedFitResults, FitResults
from hotspot3.config import ProcessorConfig

from hotspot3.connectors.bottleneck import BottleneckWrapper
from hotspot3.connectors.bam2bed import BamFileCutsExtractor

from hotspot3.io.logging import WithLoggerAndInterval, WithLogger
from hotspot3.io import check_partition_exists


class ChromReader(WithLoggerAndInterval):
    def __init__(self, genomic_interval: GenomicInterval, config: ProcessorConfig=None, logger=None):
        super().__init__(genomic_interval=genomic_interval, config=config, logger=logger)

        self.chrom_name = genomic_interval.chrom
        self.chrom_size = genomic_interval.end

        self.bn_wrapper = self.copy_with_params(BottleneckWrapper)
        self.read_extractor = self.copy_with_params(BamFileCutsExtractor)

    def extract_mappable_bases(self, mappable_file) -> np.ndarray:
        """
        Extract mappable bases for the chromosome.
        """
        if mappable_file is None:
            mappable = np.ones(self.chrom_size, dtype=bool)
        else:
            mappable = np.zeros(self.chrom_size, dtype=bool)
            try:
                with TabixExtractor(mappable_file, columns=['#chr', 'start', 'end']) as mappable_loader:
                    for _, row in mappable_loader[self.genomic_interval].iterrows():
                        if row['end'] > self.genomic_interval.end:
                            raise ValueError(f"Mappable bases file does not match chromosome sizes! Check input parameters. {row['end']} > {self.genomic_interval.end} for {self.chrom_name}")
                        mappable[row['start']:row['end']] = True
            except ValueError:
                raise NotEnoughDataForContig
    
        return mappable
    
    def extract_cutcounts(self, cutcounts_file):
        cutcounts = np.zeros(self.chrom_size, dtype=np.int32)
        try:
            with TabixExtractor(cutcounts_file) as cutcounts_loader:
                data = cutcounts_loader[self.genomic_interval]
                assert data.eval('end - start == 1').all(), "Cutcounts are expected to be at basepair resolution"
                cutcounts[data['start']] = data['count'].to_numpy()
        except ValueError:
            raise NotEnoughDataForContig

        return cutcounts
    

    def extract_aggregated_cutcounts(self, cutcounts_file):
        cutcounts = self.extract_cutcounts(cutcounts_file).astype(np.float32)
        window = self.config.window
        
        agg_cutcounts = self.bn_wrapper.centered_running_nansum(cutcounts, window)
        assert not np.any(np.isinf(agg_cutcounts)), "Aggregated cutcounts contain infinite values (in float32)!"
        return agg_cutcounts


    def extract_cutcounts_for_chrom(self, bam_path, reference_fasta=None):
        return self.read_extractor.extract_chromosome_to_df(bam_path, self.chrom_name, reference_fasta)
    

    def extract_mappable_agg_cutcounts(self, cutcounts_file, mappable_file) -> ma.MaskedArray:
        agg_cutcounts = self.extract_aggregated_cutcounts(cutcounts_file)
        mappable = self.extract_mappable_bases(mappable_file)
        return ma.masked_where(~mappable, agg_cutcounts)

    
    def extract_from_parquet(self, signal_parquet, columns) -> pd.DataFrame:
        with ChromParquetExtractor(
            signal_parquet,
            columns=columns
        ) as smoothed_signal_loader:
            signal_df = smoothed_signal_loader[self.genomic_interval]
        if signal_df.empty:
            raise NotEnoughDataForContig
        return signal_df
    
    def extract_fit_params(self, fit_params_parquet) -> WindowedFitResults:
        fit_params = self.extract_from_parquet(
            fit_params_parquet,
            columns=['sliding_r', 'sliding_p', 'enough_bg']
        )
        return WindowedFitResults(
            fit_params['sliding_p'].values,
            fit_params['sliding_r'].values,
            fit_params['enough_bg'].values
        )

    def extract_trs(self, fit_params_parquet) -> np.ndarray:
        trs = self.extract_from_parquet(
            fit_params_parquet,
            columns=['tr']
        )
        return trs['tr'].values

    def read_and_update_fit_params(self, fit_params_parquet, fit_results: FitResults):
        fit_params = self.extract_fit_params(fit_params_parquet)
        fit_results.p[fit_params.enough_bg_mask] = fit_params.p[fit_params.enough_bg_mask]
        fit_results.r[fit_params.enough_bg_mask] = fit_params.r[fit_params.enough_bg_mask]
        return fit_results
    
    def extract_fit_threholds(self, fit_params_parquet):
        threholds = self.extract_from_parquet(
            fit_params_parquet,
            columns=['chrom', 'tr']
        )
        return threholds
    
    def extract_significant_bases(self, fdrs, fdr_threshold, min_signif_bases=1):
        signif_bases = fdrs >= -np.log10(fdr_threshold)
        smoothed_signif = self.bn_wrapper.centered_running_nansum(
            signif_bases,
            window=self.config.window,
        ) >= min_signif_bases
        return smoothed_signif


class GenomeReader(WithLogger):

    def read_full_parquet(self, pvals_path, column, **kwargs):
        return pd.read_parquet(
            pvals_path,
            engine='pyarrow', 
            columns=[column],
            **kwargs
        )[column]

    def read_pval_from_parquet(self, pvals_path, **kwargs):
        return pd.read_parquet(pvals_path, engine='pyarrow', columns=['log10_pval'], **kwargs)['log10_pval'].values
    
    def read_chrom_pos_mapping(
        self,
        pvals_path,
        chrom_sizes: dict=None,
    ) -> List[GenomicInterval]:
        if chrom_sizes is not None:
            chroms = sorted([
                x for x in chrom_sizes.keys() 
                if check_partition_exists(pvals_path, ['chrom'], [x])
            ])
            chrom_sizes = [0] + [chrom_sizes[y] for y in chroms]
            index_of_chrom = np.cumsum(chrom_sizes)
            starts = index_of_chrom[:-1]
            ends = index_of_chrom[1:]
        else: 
            chroms = self.read_full_parquet(pvals_path, column='chrom')
            total_len = chroms.shape[0]
            chroms = chroms.drop_duplicates()
            starts = chroms.index
            # file is always sorted within chromosomes
            ends = [*starts[1:], total_len]
        return [GenomicInterval(chrom, start, end) for chrom, start, end in zip(chroms, starts, ends)]
    
    
    def get_chrom_sizes_file(self, chrom_sizes_file):
        if chrom_sizes_file is None:
            raise NotImplementedError("Chromosome sizes file is not embedded yet")
        return chrom_sizes_file

    def read_chrom_sizes(self, chrom_sizes):
        return pd.read_table(
            self.get_chrom_sizes_file(chrom_sizes),
            header=None,
            names=['chrom', 'size']
        ).set_index('chrom')['size'].to_dict()

    def extract_cutcounts_all_chroms(self, bam_path, tabix_path, chromosomes, reference_fasta=None):
        read_extractor = self.copy_with_params(BamFileCutsExtractor)
        return read_extractor.extract_all_chroms(bam_path, tabix_path, chromosomes, reference_fasta)
    
    def read_total_cutcounts(self, total_cutcounts_path):
        return np.loadtxt(total_cutcounts_path, dtype=int)



import pandas as pd
import os
import shutil
import tempfile
from typing import List


from hotspot3.helpers.models import NotEnoughDataForContig


def to_parquet_high_compression(df: pd.DataFrame, outpath, compression_level=22, partition_cols=None, **kwargs):
    if partition_cols is None:
        partition_cols = ['chrom']

    df.to_parquet(
        outpath,
        engine='pyarrow',
        compression='zstd',
        use_dictionary=True,
        index=False,
        partition_cols=partition_cols,
        compression_level=compression_level,
        **kwargs
    )


def parallel_write_partitioned_parquet(
        data_df,
        field_names,
        partition_cols,
        path,
        tmp_dir=None,
        *args,
        **kwargs
    ):
    assert len(field_names) == len(partition_cols)
    os.makedirs(path, exist_ok=True)
    with tempfile.TemporaryDirectory(dir=tmp_dir) as temp_dir:
        tmp_prefix = '.'.join(field_names)
        temp_path = os.path.join(temp_dir, f'{tmp_prefix}.temp.parquet')
        to_parquet_high_compression(
            data_df,
            temp_path,
            partition_cols=partition_cols,
            *args,
            **kwargs
        )
        src, dest = get_src_and_dest_partioned_parquet(temp_path, path, partition_cols, field_names)
        dest_dir = os.path.dirname(dest)
        if os.path.exists(dest_dir):
            shutil.rmtree(dest_dir)
        os.makedirs(dest_dir, exist_ok=True)
        shutil.move(src, dest)


def get_src_and_dest_partioned_parquet(src, dest, partition_cols, field_names):
    names = [
        f'{partition_col}={name}' 
        for partition_col, name in zip(partition_cols, field_names)
    ]
    parquet_file = [x for x in os.listdir(os.path.join(src, *names)) if x.endswith('.parquet')]
    assert len(parquet_file) == 1
    parquet_file = parquet_file[0]
    return os.path.join(src, *names, parquet_file), os.path.join(dest, *names, parquet_file)


def read_partioned_parquet(filename, partition_cols, partition_vals, columns=None):
    assert len(partition_cols) == len(partition_vals)
    if not check_partition_exists(filename, partition_cols, partition_vals):
        raise NotEnoughDataForContig
    return pd.read_parquet(
            filename,
            filters=[
                (partition_col, '==', partition_val) 
                for partition_col, partition_val in zip(partition_cols, partition_vals)
            ],
            engine='pyarrow',
            columns=columns
        )


def check_partition_exists(parquet_root: str, partition_cols: List[str], partition_vals: List[str]):
    """
    Check if a specific partition exists in a Parquet dataset.

    Parameters:
        parquet_root (str): Path to the root of the Parquet dataset.
        partition_cols (List[str]): List of partition column names.
        partition_vals (List[str]): List of partition values.

    Returns:
        bool: True if the partition exists, False otherwise.
    """
    target_path = os.path.join(parquet_root, *[
        f"{col}={val}" for col, val in zip(partition_cols, partition_vals)
    ])
    return os.path.exists(target_path)



import pandas as pd
import argparse
import numpy as np
import sys
from tqdm import tqdm

from hotspot3.io.readers import GenomeReader, ChromReader
from hotspot3.helpers.models import NotEnoughDataForContig
from genome_tools import GenomicInterval

try:
    import numba

except ImportError:
    print("Numba is not installed. Please install it to use this script.")
    sys.exit(1)


@numba.njit
def segment_max(vec, starts, ends, out):
    for i in range(starts.shape[0]):
        m = vec[starts[i]]
        for j in range(starts[i]+1, ends[i]+1):
            if vec[j] > m:
                m = vec[j]
        out[i] = m


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('pvals_parquet', type=str, help='Path to parquet file with pvals')
    parser.add_argument('bed', type=str, help='Path to bed file with regions to extract pvals from. Use "-" for stdin')
    parser.add_argument('--format', choices=['bed', 'txt', 'npy'], default='bed', help='Output format')
    parser.add_argument("--chrom_sizes", help="Path to chromosome sizes file. If none assumed to be hg38 sizes", default=None)

    parser.add_argument('save_path', type=str, help='Path to save results')

    return parser.parse_args()

def extract_max_pval(row, chrom_pvals):
    return np.nanmax(chrom_pvals[row['start']:row['end']], initial=0)

def main():
    args = parse_args()
    reader = GenomeReader()
    chrom_sizes = reader.read_chrom_sizes(args.chrom_sizes)
    if args.bed == '-':
        input_bed = sys.stdin
    else:
        input_bed = args.bed
    bed_df = pd.read_table(input_bed, header=None, names=['chrom', 'start', 'end'], comment='#', usecols=[0, 1, 2])
    groups = bed_df.groupby('chrom')
    data = []
    for chrom, group in tqdm(groups, total=len(groups)):
        if chrom not in chrom_sizes:
            max_pvals = np.zeros(len(group), dtype=np.float32)
        else:
            chrom_interval = GenomicInterval(chrom, 0, chrom_sizes[chrom])

            chrom_reader = ChromReader(genomic_interval=chrom_interval)
            try:
                chrom_pvals = chrom_reader.extract_from_parquet(
                    args.pvals_parquet,
                    columns=['log10_pval']
                )['log10_pval'].values
                max_pvals = group.apply(extract_max_pval, chrom_pvals=chrom_pvals, axis=1)
            except NotEnoughDataForContig:
                max_pvals = np.zeros(len(group), dtype=np.float32)
            
        group['max_neglog_p'] = max_pvals
        data.append(group)

    data = pd.concat(data).sort_index()
    if args.format == 'npy':
        np.save(args.save_path, data['max_neglog_p'].astype(np.float32).values)
    elif args.format == "txt":
        np.savetxt(
            args.save_path,
            data['max_neglog_p'].astype(np.float32).values,
        )
    else:
        data.to_csv(
            args.save_path,
            sep='\t',
            index=False,
            header=None
        )

if __name__ == '__main__':
    main()

        





import numpy as np
from concurrent.futures import ProcessPoolExecutor
import multiprocessing as mp
import pandas as pd
from typing import Iterable, List, cast
import functools

from genome_tools.genomic_interval import GenomicInterval, df_to_genomic_intervals

from hotspot3.helpers.models import ProcessorOutputData, NotEnoughDataForContig
from hotspot3.helpers.utils import is_iterable, ensure_contig_exists
from hotspot3.helpers.stats import upper_bg_quantile
from hotspot3.helpers.format_converters import fit_stats_df_to_fallback_fit_results, peaks_to_bed12, hotspots_to_bed12, fit_results_to_df

from hotspot3.io.logging import WithLoggerAndInterval, WithLogger
from hotspot3.io.readers import ChromReader, GenomeReader
from hotspot3.io.writers import ChromWriter, GenomeWriter

from hotspot3.connectors.babachi import BabachiWrapper

from hotspot3.signal_smoothing import modwt_smooth, normalize_density
from hotspot3.background_fit.genomic_background_fit import ChromosomeFit, SegmentalFit
from hotspot3.background_fit.regression import SignalToNoiseFit
from hotspot3.scoring.pvalue import PvalueEstimator
from hotspot3.scoring.fdr import SampleFDRCorrection
from hotspot3.peak_calling import find_stretches, find_varwidth_peaks


def parallel_func_error_handler(func):
    @functools.wraps(func)
    def wrapper(processor: 'ChromosomeProcessor', *args, **kwargs):
        try:
            return func(processor, *args, **kwargs)
        except:
            processor.logger.exception(f"Exception occured in {func.__name__} for chromosome {processor.chrom_name}")
            raise
    return wrapper


class GenomeProcessor(WithLogger):
    """
    Main class to run hotspot3-related functions and store parameters.

    Parameters:
        - chrom_sizes: Path to chromosome sizes file.
        - mappable_bases_file: Path to the tabix-indexed file containing mappable bases or None.
        - tmp_dir: Temporary directory for intermediate files. Will use system default if None.
        - chromosomes: List of chromosomes to process or None. Used mostly for debugging. Will generate FDR corrections only for these chromosomes.

        - config: ProcessorConfig object containing parameters.
    """
    def __init__(
        self,
        sample_id,
        chrom_sizes_file=None,
        config=None,
        mappable_bases_file=None, 
        reference_fasta=None,
        chromosomes=None
    ):
        super().__init__(config=config)
        self.sample_id = sample_id
        self.mappable_bases_file = mappable_bases_file
        self.chrom_sizes_file = chrom_sizes_file
        self.reader = self.copy_with_params(GenomeReader)
        self.reference_fasta = reference_fasta

        self.cpus = min(self.config.cpus, max(1, mp.cpu_count()))

        self.chrom_sizes = self.reader.read_chrom_sizes(chrom_sizes_file)

        if chromosomes is not None:
            self.chrom_sizes = {k: v for k, v in self.chrom_sizes.items() if k in chromosomes}
        
        chroms = [x for x in self.chrom_sizes.keys()]
        self.logger.debug(f"Chromosomes to process: {chroms}")

        chrom_intervals = [
            GenomicInterval(chrom, 0, length) 
            for chrom, length in self.chrom_sizes.items()
        ]
        self.chromosome_processors = sorted(
            [
                ChromosomeProcessor(self, interval, config=self.config, logger=self.logger) 
                for interval in chrom_intervals
            ],
            key=lambda x: x.chrom_size,
            reverse=True
        )
        self.writer = self.copy_with_params(GenomeWriter)
    
    ## Parallel processing functions

    def construct_parallel_args(self, *args):
        res_args = []
        for arg in args:
            if is_iterable(arg):
                if all(isinstance(x, ProcessorOutputData) for x in arg):
                     # if arg consits of ProcessorOutputData - 
                     # sort by chromosome name to match chromosome_processors
                    arg = cast(List[ProcessorOutputData], arg)
                    tmp = {x.id: x for x in arg}
                    reformat_arg = []
                    for x in self.chromosome_processors:
                        if x.chrom_name in tmp:
                            d = tmp[x.chrom_name]
                        else:
                            self.logger.debug(f"Chromosome {x.chrom_name} not found in input data. Skipping.")
                            d = None
                        reformat_arg.append(d)

                else:
                    assert len(arg) == len(self.chromosome_processors), f"Length of arguments must be equal to the number of chromosomes ({len(self.chromosome_processors)})."
                    reformat_arg = arg
            else:
                reformat_arg = [arg] * len(self.chromosome_processors)
            res_args.append(reformat_arg)
        return [self.chromosome_processors, *res_args]


    def parallel_by_chromosome(self, func, *args, cpus=None):
        """
        Basic function that handles parallel execution of a function by chromosome.
        """
        if cpus is None: # override cpus if provided
            cpus = self.cpus
        args = self.construct_parallel_args(*args)
        self.logger.debug(f'Using {cpus} CPUs to {func.__name__}')
        results = []
        if self.cpus == 1:
            for func_args in zip(*args):
                result = func(*func_args)
                if result is not None:
                    results.append(result)
        else:
            with ProcessPoolExecutor(max_workers=self.cpus) as executor:
                try:
                    for result in executor.map(func, *args):
                        if result is not None:
                            results.append(result)
                except Exception as e:
                    self.logger.critical(f"Exception occured, gracefully shutting down executor...")
                    self.logger.critical(e)
                    raise e

        self.logger.debug(f'Results of {func.__name__} emitted.')
        return results
    
    def merge_and_add_chromosome(self, results: Iterable[ProcessorOutputData]) -> ProcessorOutputData:
        data = []
        results = {r.id: r for r in results}
        results: List[ProcessorOutputData] = [
            results[chrom] for chrom in self.chrom_sizes if chrom in results
        ]
        categories = [x.id for x in results]
        for res in results:
            df = res.data_df
            df['chrom'] = pd.Categorical(
                [res.id] * df.shape[0],
                categories=categories,
            )
            data.append(df)
        
        if len(data) == 0:
            return ProcessorOutputData('all', pd.DataFrame(columns=['chrom']))
        data = pd.concat(data, ignore_index=True)
        data = data[['chrom'] + [col for col in data.columns if col != 'chrom']]
        return ProcessorOutputData('all', data)


    ## Processing functions
    def extract_cutcounts_from_bam(self, bam_path, outpath):
        """
        Extract cutcounts from BAM file and save to tabix file.

        Parameters:
            - bam_path: Path to the BAM/CRAM/SAM file.
            - outpath: Path to save the cutcounts to.
        """
        self.logger.info('Extracting cutcounts from BAM/CRAM file')
        if False: # self.cpus >= 10: # FIXME make it work for large cpu (now concurent read to cram make the script fail for some reason)
            data = self.parallel_by_chromosome(
                ChromosomeProcessor.extract_cutcounts_for_chromosome,
                bam_path,
            )
            data = self.merge_and_add_chromosome(data).data_df
            self.writer.df_to_tabix(data, outpath)
        else:
            # BAM2BED writes to file faster
            self.reader.extract_cutcounts_all_chroms(
                bam_path,
                outpath,
                self.chrom_sizes.keys(),
                reference_fasta=self.reference_fasta
            )


    def get_total_cutcounts(self, cutcounts_path, total_cutcounts_path) -> int:
        """
        Extract total cutcounts from cutcounts file and save to a separate file.

        Parameters:
            - cutcounts_path: Path to tabix file containing the cutcounts.
            - total_cutcounts_path: Path to save the total cutcounts to.
        """
        total_cutcounts = sum(
            self.parallel_by_chromosome(
                ChromosomeProcessor.get_total_cutcounts,
                cutcounts_path
            )
        )
        self.logger.info('Total cutcounts = %d', total_cutcounts)
        if total_cutcounts == 0:
            self.logger.critical('Total # of cuts (ends of reads) is 0. Most likely the input is malformed or empty. Exiting.')
            raise ValueError('Total # of cuts (ends of reads) is 0. Most likely the input is malformed or empty. Exiting.')
        self.writer.save_cutcounts(total_cutcounts, total_cutcounts_path)


    def smooth_signal_modwt(self, cutcounts_path, save_path, total_cutcounts_path):
        """
        Smooth signal using MODWT and save to parquet file.
        """
        self.logger.info('Smoothing signal using MODWT')
        total_cutcounts = self.reader.read_total_cutcounts(total_cutcounts_path)
        
        self.writer.sanitize_path(save_path)
        self.parallel_by_chromosome(
            ChromosomeProcessor.smooth_density_modwt,
            cutcounts_path,
            total_cutcounts,
            save_path
        )

    def extract_fit_thresholds_to_bw(self, fit_params_path, total_cutcounts_path, save_path):
        """
        Convert threshold values to BigWig file.
        """
        self.logger.info('Converting threshold values to BigWig')
        thresholds = self.parallel_by_chromosome(
            ChromosomeProcessor.extract_fit_threholds,
            fit_params_path
        )
        total_cutcounts = self.reader.read_total_cutcounts(total_cutcounts_path)
        thresholds = self.merge_and_add_chromosome(thresholds).data_df

        self.writer.thresholds_df_to_bw(
            thresholds,
            save_path,
            total_cutcounts,
            chrom_sizes=self.chrom_sizes
        )
    
    def extract_bg_quantile_to_bw(self, fit_params_path, total_cutcounts_path, save_path):
        """
        Convert threshold values to BigWig file.
        """
        self.logger.info('Converting background quantile values to BigWig')
        thresholds = self.parallel_by_chromosome(
            ChromosomeProcessor.extract_bg_quantile,
            fit_params_path
        )
        total_cutcounts = self.reader.read_total_cutcounts(total_cutcounts_path)
        thresholds = self.merge_and_add_chromosome(thresholds).data_df

        self.writer.thresholds_df_to_bw(
            thresholds,
            save_path,
            total_cutcounts,
            chrom_sizes=self.chrom_sizes
        )

    def refit_outlier_segments(self, cutcounts_file, params_df: pd.DataFrame):
        outlier_params = params_df.query('refit_with_constraint | fit_type == "global"')

        bad_segments = [
            ProcessorOutputData(x[0], x[1]) 
            for x in outlier_params.groupby('chrom', observed=True)
        ]

        refit_params = self.parallel_by_chromosome(
            ChromosomeProcessor.refit_outlier_segments,
            cutcounts_file,
            bad_segments
        )
        refit_params = self.merge_and_add_chromosome(refit_params).data_df
        return refit_params


    def fit_background_model(
            self,
            cutcounts_file,
            total_cutcounts_path,
            save_path: str,
            per_region_stats_path,
            per_region_stats_path_bw
        ):
        """"
        Fit background model of cutcounts distribution and save fit parameters to parquet file.
        """
        self.logger.info('Estimating parameters of the background model')
        
        per_region_params = self.parallel_by_chromosome(
            ChromosomeProcessor.fit_background_model,
            cutcounts_file,
        )
        per_region_params = self.merge_and_add_chromosome(per_region_params).data_df

        sn_fit = self.copy_with_params(SignalToNoiseFit)
        per_region_params, spot_results = sn_fit.fit(per_region_params)
        per_region_params['refit_with_constraint'] = sn_fit.find_outliers(per_region_params)

        # keep track of all segments that have been refitted through the iterations
        refit_with_constraint = per_region_params['refit_with_constraint'].values

        for iteration in range(1, self.config.max_outlier_iterations + 1):
            is_outlier_segment = per_region_params['refit_with_constraint']
            if is_outlier_segment.sum() == 0:
                break

            self.logger.info(f"Found {is_outlier_segment.sum()} outlier SPOT score segments at iteration {iteration}. Refitting with approximated signal/noise constraint.")

            if self.config.save_debug:
                self.writer.df_to_tabix(per_region_params, per_region_stats_path + f'.iter{iteration}')

            refit_params = self.refit_outlier_segments(
                cutcounts_file=cutcounts_file,
                params_df=per_region_params
            )
            per_region_params.loc[is_outlier_segment, refit_params.columns] = refit_params.values

            # Refit the model and check for outliers
            per_region_params, spot_results = sn_fit.fit(per_region_params)
            per_region_params['refit_with_constraint'] = sn_fit.find_outliers(per_region_params)

            refit_with_constraint |= per_region_params['refit_with_constraint']
        else:
            self.logger.info(f'No outlier segments found at iteration {iteration}.')

        per_region_params['refit_with_constraint'] = refit_with_constraint
        per_region_params['valid_segment'] = per_region_params.eval('success_fit & ~max_bg_reached & fit_type == "segment"')

        self.logger.info(f"Final SPOT score: {spot_results.spot_score:.2f}±{spot_results.spot_score_std:.2f}. Refitted {refit_with_constraint.sum()} segments.")
        self.writer.df_to_tabix(per_region_params, per_region_stats_path)
        self.writer.fit_stats_to_bw(
            per_region_params,
            per_region_stats_path_bw,
            chrom_sizes=self.chrom_sizes,
            total_cutcounts=self.reader.read_total_cutcounts(total_cutcounts_path),
        )
        self.logger.info('Estimating per-bp parameters of the background model')
        self.writer.sanitize_path(save_path)
        all_segments = [
            ProcessorOutputData(x[0], x[1]) 
            for x in per_region_params.query('fit_type == "segment"').groupby('chrom', observed=True)
        ]
        self.parallel_by_chromosome(
            ChromosomeProcessor.fit_per_bp_model,
            cutcounts_file,
            all_segments,
            save_path
        )

    def calc_pvals(self, cutcounts_file, fit_path: str, save_path: str):
        """
        Calculate per-bp p-values from cutcounts and background model parameters.
        """
        self.logger.info('Calculating per-bp p-values')
        
        self.writer.sanitize_path(save_path)
        self.parallel_by_chromosome(
            ChromosomeProcessor.calc_pvals,
            cutcounts_file,
            fit_path,
            save_path
        )

    
    def calc_fdr(self, pvals_path, save_path, max_fdr=1):
        """
        Calculate fast per-bp FDRs from p-values. If max_fdr is set to 1, will calculate all FDRs.

        Parameters:
            - pvals_path: Path to the parquet file containing the p-values.
            - save_path: Path to save the FDRs in parquet format.
            - max_fdr: Maximum FDR to calculate.
        """
        self.logger.info('Per-bp FDR correction')
        self.copy_with_params(
            SampleFDRCorrection,
            name=self.sample_id,
            chrom_sizes=self.chrom_sizes
        ).fdr_correct_pvals(
            pvals_path,
            max_fdr,
            save_path,
        )

    def call_hotspots(self, fdr_path, sample_id, save_path, save_path_bb, fdr_tr):
        """
        Call hotspots from parquet file containing log10(FDR) values.

        Parameters:
            - fdr_path: Path to the parquet file containing the log10(FDR) values.
            - sample_id: Sample ID.
            - save_path: Path to save the hotspots in bed format.
            - save_path_bb: Path to save the hotspots in bigbed format.
            - fdr_tr: FDR threshold for calling hotspots.
        """
        hotspots = self.parallel_by_chromosome(
            ChromosomeProcessor.call_hotspots,
            fdr_path,
            fdr_tr,
        )
        hotspots = self.merge_and_add_chromosome(hotspots).data_df
        signif_stretches = hotspots['signif_stretches'].values
        hotspots['id'] = sample_id
        if len(hotspots) == 0:
            self.logger.critical(f"No hotspots called at FDR={fdr_tr}. Most likely something went wrong!")
        else:
            self.logger.info(f"There are {len(hotspots)} hotspots called at FDR={fdr_tr}")
        
        hotspots = hotspots[['chrom', 'start', 'end', 'id', 'max_neglog10_fdr']]
        self.writer.df_to_tabix(hotspots, save_path)

        hotspots = hotspots_to_bed12(hotspots, fdr_tr, signif_stretches)
        self.writer.df_to_bigbed(hotspots, self.chrom_sizes_file, save_path_bb)


    def call_variable_width_peaks(
            self,
            smoothed_signal_path,
            fdrs_path,
            total_cutcounts_path,
            sample_id,
            save_path,
            save_path_bb,
            fdr_tr=0.05
        ):
        """
        Call variable width peaks from smoothed signal and parquet file containing log10(FDR) values.

        Parameters:
            - smoothed_signal_path: Path to the parquet file containing the smoothed signal.
            - fdrs_path: Path to the parquet file containing the log10(FDR) values.
            - sample_id: Sample ID.
            - save_path: Path to save the peaks in bed format.
            - save_path_bb: Path to save the peaks in bigbed format.
            - fdr_tr: FDR threshold for calling peaks.
        """
        peaks_data = self.parallel_by_chromosome(
            ChromosomeProcessor.call_variable_width_peaks,
            smoothed_signal_path,
            fdrs_path,
            fdr_tr
        )
        total_cutcounts = self.reader.read_total_cutcounts(total_cutcounts_path)
        peaks = self.merge_and_add_chromosome(peaks_data).data_df
        if len(peaks) == 0:
            self.logger.critical(f"No peaks called at FDR={fdr_tr}. Most likely something went wrong!")
            return
        else:
            self.logger.info(f"There are {len(peaks)} peaks called at FDR={fdr_tr}")
        
        peaks['id'] = sample_id
        peaks = peaks[['chrom', 'start', 'end', 'id', 'max_density', 'summit_density', 'summit', 'smoothed_peak_height']]
        peaks['smoothed_peak_height'] = normalize_density(
            peaks['smoothed_peak_height'],
            total_cutcounts
        )
        self.writer.df_to_tabix(peaks, save_path)

        peaks = peaks_to_bed12(peaks, fdr_tr)
        self.writer.df_to_bigbed(peaks, self.chrom_sizes_file, save_path_bb)


    def extract_normalized_density(self, smoothed_signal, density_path):
        """
        Save normalized density to BigWig file. Uses the density_step parameter from the config to downsample the data.

        Parameters:
            - smoothed_signal: Path to the parquet file containing the smoothed signal.
            - density_path: Path to save the BigWig file.
        """
        density_data = self.parallel_by_chromosome(
            ChromosomeProcessor.extract_normalized_density,
            smoothed_signal
        )
        density_data = self.merge_and_add_chromosome(density_data).data_df
        self.writer.density_to_bw(
            density_data,
            density_path,
            chrom_sizes=self.chrom_sizes
        )
        

class ChromosomeProcessor(WithLoggerAndInterval):
    """
    Individual chromosome processor. Used for parallel processing of chromosomes.
    Don't use directly, use GenomeProcessor instead.
    """
    def __init__(self, genome_processor: GenomeProcessor, chromosome_interval: GenomicInterval, config=None, logger=None) :
        super().__init__(chromosome_interval, config=config, logger=logger)
        self.chrom_name = self.genomic_interval.chrom
        self.gp = genome_processor
        
        self.chrom_size = len(self.genomic_interval)
        self.reader = self.copy_with_params(ChromReader)
        self.writer = self.copy_with_params(ChromWriter)


    @parallel_func_error_handler
    @ensure_contig_exists
    def extract_cutcounts_for_chromosome(self, bam_path) -> ProcessorOutputData:
        data = self.reader.extract_cutcounts_for_chrom(bam_path, self.gp.reference_fasta)
        return ProcessorOutputData(self.chrom_name, data)


    @parallel_func_error_handler
    @ensure_contig_exists
    def get_total_cutcounts(self, cutcounts) -> int:
        """
        Get total # of cutcounts for the chromosome.
        """
        self.logger.debug(f"{self.chrom_name}: Calculating total cutcounts")
        return self.reader.extract_cutcounts(cutcounts).sum()


    @parallel_func_error_handler
    @ensure_contig_exists
    def smooth_density_modwt(self, cutcounts_path, total_cutcounts, save_path):
        """
        Run MODWT smoothing on cutcounts. Smoothed density will be used to call peaks.
        """
        self.logger.debug(f"{self.chrom_name}: Running modwt signal smoothing (filter={self.config.filter}, level={self.config.modwt_level})")
        agg_cutcounts = self.reader.extract_aggregated_cutcounts(cutcounts_path)
        smoothed = modwt_smooth(agg_cutcounts, self.config.filter, level=self.config.modwt_level)
        data = pd.DataFrame({
            'smoothed': smoothed,
            'normalized_density': normalize_density(agg_cutcounts, total_cutcounts) 
        })
        self.write_to_parquet(data, save_path)
    

    @parallel_func_error_handler
    @ensure_contig_exists
    def fit_background_model(self, cutcounts_file) -> ProcessorOutputData:
        """
        Fit background model to cutcounts and save fit parameters to a bed.gz file.
        """
        agg_cutcounts = self.reader.extract_mappable_agg_cutcounts(
            cutcounts_file,
            self.gp.mappable_bases_file
        )
        self.logger.debug(f'{self.chrom_name}: Estimating proportion of background vs signal')
        
        chrom_fit = self.copy_with_params(ChromosomeFit)
        per_window_trs_global, global_fit_params = chrom_fit.fit_segment_thresholds(agg_cutcounts)
        
        if global_fit_params.rmsea > self.config.rmsea_tr:
            self.logger.warning(f"{self.chrom_name}: Best RMSEA: {global_fit_params.rmsea:.3f}. Chromosome fit might be poorly approximated.")

        self.logger.debug(f"{self.chrom_name}: Signal quantile: {global_fit_params.fit_quantile:.3f}. signal threshold: {global_fit_params.fit_threshold:.0f}. Best RMSEA: {global_fit_params.rmsea:.3f}")

        # Segmentation
        segmentation = self.copy_with_params(BabachiWrapper)
        bad_segments = segmentation.run_segmentation(
            agg_cutcounts,
            per_window_trs_global,
            global_fit_params
        )
        segments_fit = self.copy_with_params(SegmentalFit)
        per_interval_params = segments_fit.fit_per_segment_bg_model(
            agg_cutcounts=agg_cutcounts,
            bad_segments=bad_segments,
            fallback_fit_results=global_fit_params
        )
        per_interval_params = segments_fit.add_fallback_fit_stats(
            global_fit_params,
            per_interval_params,
        )
        
        return ProcessorOutputData(self.chrom_name, per_interval_params)


    @parallel_func_error_handler
    @ensure_contig_exists
    def refit_outlier_segments(
        self,
        cutcounts_path,
        bad_segments: ProcessorOutputData,
    ) -> ProcessorOutputData:
        if bad_segments is None:
            raise NotEnoughDataForContig
        segments = bad_segments.data_df.query(f'fit_type == "segment"')
        if segments.empty:
            raise NotEnoughDataForContig
        
        segment_intervals = df_to_genomic_intervals(segments, extra_columns=['BAD'])
        chrom_fit = fit_stats_df_to_fallback_fit_results(
            bad_segments.data_df.query('fit_type == "global"')
        )
        agg_cutcounts = self.reader.extract_mappable_agg_cutcounts(
            cutcounts_path,
            self.gp.mappable_bases_file
        )
        segments_fit = self.copy_with_params(SegmentalFit)
        per_interval_params = segments_fit.fit_per_segment_bg_model(
            agg_cutcounts=agg_cutcounts,
            bad_segments=segment_intervals,
            fallback_fit_results=chrom_fit,
            min_bg_tag_proportion=segments['min_bg_tag_proportion'].values
        )

        return ProcessorOutputData(self.chrom_name, per_interval_params)
    

    @parallel_func_error_handler
    @ensure_contig_exists
    def fit_per_bp_model(
        self,
        cutcounts_path,
        segment_fit_results: ProcessorOutputData,
        save_path
    ):
        if segment_fit_results is None:
            raise NotEnoughDataForContig
        segments_fit = self.copy_with_params(SegmentalFit)
        agg_cutcounts = self.reader.extract_mappable_agg_cutcounts(
            cutcounts_path,
            self.gp.mappable_bases_file
        )
        per_bp_params, fit_trs = segments_fit.per_bp_background_model_fit(
            agg_cutcounts,
            segment_fit_results.data_df
        )
        df = fit_results_to_df(per_bp_params, fit_trs)
        self.write_to_parquet(df, save_path)


    @parallel_func_error_handler
    @ensure_contig_exists
    def extract_fit_threholds(self, fit_parquet_path) -> ProcessorOutputData:
        fit_res = self.reader.extract_fit_threholds(fit_parquet_path).iloc[::self.config.bg_track_step]
        fit_res['start'] = np.arange(len(fit_res)) * self.config.bg_track_step
        fit_res.dropna(inplace=True)
        return ProcessorOutputData(self.chrom_name, fit_res)


    @parallel_func_error_handler
    @ensure_contig_exists
    def extract_bg_quantile(self, fit_parquet_path) -> ProcessorOutputData:
        fit_res = self.reader.extract_fit_params(fit_parquet_path)
        fit_res = upper_bg_quantile(
            fit_res.r[::self.config.bg_track_step],
            fit_res.p[::self.config.bg_track_step]
        )
        fit_res = pd.DataFrame({
            'tr': fit_res,
            'start': np.arange(len(fit_res)) * self.config.bg_track_step
        }).dropna()
        return ProcessorOutputData(self.chrom_name, fit_res)

    @parallel_func_error_handler
    @ensure_contig_exists
    def calc_pvals(self, cutcounts_file, fit_parquet_path, save_path) -> ProcessorOutputData:
        self.logger.debug(f'{self.chrom_name}: Calculating p-values')
        agg_cutcounts = self.reader.extract_mappable_agg_cutcounts(
            cutcounts_file,
            self.gp.mappable_bases_file
        )
        agg_cutcounts = np.floor(agg_cutcounts.filled(np.nan))
        fit_res = self.reader.extract_fit_params(fit_parquet_path)
    
        pval_estimator = self.copy_with_params(PvalueEstimator)

        neglog_pvals = pval_estimator.estimate_pvalues(agg_cutcounts, fit_res)
        self.logger.debug(f"Saving p-values for {self.chrom_name}")
 
        neglog_pvals = pd.DataFrame({'log10_pval': pval_estimator.fix_inf_pvals(neglog_pvals)})
        self.write_to_parquet(neglog_pvals, save_path)


    @parallel_func_error_handler
    @ensure_contig_exists
    def call_hotspots(self, fdr_path, fdr_threshold=0.05) -> ProcessorOutputData:
        
        log10_fdrs = self.copy_with_params(
            SampleFDRCorrection,
            name=self.gp.sample_id
        ).read_fdrs_for_chrom(fdr_path, self.chrom_name)
        self.logger.debug(f"{self.chrom_name}: Calling hotspots")
        
        # TODO: move logic to a peak calling class
        signif_fdrs = (log10_fdrs >= -np.log10(fdr_threshold))
        smoothed_signif = self.reader.extract_significant_bases(
            log10_fdrs,
            fdr_threshold,
            min_signif_bases=1
        )
        region_starts, region_ends = find_stretches(smoothed_signif)

        max_log10_fdrs = np.empty(region_ends.shape)
        signif_stretches = []
        n_signif_bases = []
        for i in range(len(region_starts)):
            start = region_starts[i]
            end = region_ends[i]
            max_log10_fdrs[i] = np.nanmax(log10_fdrs[start:end])
            signif_stretches.append(find_stretches(signif_fdrs[start:end]))
            n_signif_bases.append(np.sum(signif_fdrs[start:end]))

        self.logger.debug(f"{self.chrom_name}: {len(region_starts)} hotspots called")

        data = pd.DataFrame({
            'start': region_starts,
            'end': region_ends,
            'max_neglog10_fdr': max_log10_fdrs,
            'n_significant_bases': n_signif_bases,
            'signif_stretches': signif_stretches,
        })
        return ProcessorOutputData(self.chrom_name, data)


    @parallel_func_error_handler
    @ensure_contig_exists
    def call_variable_width_peaks(self, smoothed_signal_path, fdr_path, fdr_threshold) -> ProcessorOutputData:
        signal_df = self.reader.extract_from_parquet(
            smoothed_signal_path,
            columns=['smoothed', 'normalized_density']
        )
        log10_fdrs = self.copy_with_params(
            SampleFDRCorrection,
            name=self.gp.sample_id
        ).read_fdrs_for_chrom(fdr_path, self.chrom_name)

        self.logger.debug(f"{self.chrom_name}: Calling peaks at FDR={fdr_threshold}")
        
        # TODO: move logic to a peak calling class
        smoothed_signif = self.reader.extract_significant_bases(
            log10_fdrs,
            fdr_threshold,
            min_signif_bases=self.config.min_signif_bases_for_peak
        )

        starts, ends = find_stretches(smoothed_signif)
        if len(starts) == 0:
            self.logger.warning(f"{self.chrom_name}: No peaks found. Skipping...")
            raise NotEnoughDataForContig

        normalized_density = signal_df['normalized_density'].values
        peaks_in_hotspots_trimmed, cutoff_height = find_varwidth_peaks(
            signal_df['smoothed'].values,
            starts,
            ends
        )
        peaks_df = pd.DataFrame(
            peaks_in_hotspots_trimmed,
            columns=['start', 'summit', 'end']
        )
        # TODO: wrap in function
        peaks_df['end'] = peaks_df['summit'] + np.maximum(peaks_df.eval('end - summit'), self.config.min_peak_half_width)
        peaks_df['start'] = peaks_df['summit'] - np.maximum(peaks_df.eval('summit - start'), self.config.min_peak_half_width)
    
        peaks_df['end'] += 1 # BED format, half-open intervals
        peaks_df['summit_density'] = normalized_density[peaks_df['summit']]
        peaks_df['smoothed_peak_height'] = cutoff_height
        
        peaks_df['max_density'] = [
            np.max(normalized_density[start:end])
            for start, end in zip(peaks_df['start'], peaks_df['end'])
        ]
        self.logger.debug(f"{self.chrom_name}: {len(peaks_df)} peaks called")

        return ProcessorOutputData(self.chrom_name, peaks_df)


    @parallel_func_error_handler
    @ensure_contig_exists
    def extract_normalized_density(self, smoothed_signal) -> ProcessorOutputData:
        data_df = self.reader.extract_from_parquet(
            smoothed_signal, 
            columns=['chrom', 'normalized_density']
        )[::self.config.density_track_step]
        
        data_df['start'] = np.arange(len(data_df)) * self.config.density_track_step
        data_df.query('normalized_density > 0', inplace=True)
        return ProcessorOutputData(self.chrom_name, data_df)


    @parallel_func_error_handler
    @ensure_contig_exists
    def write_to_parquet(self, data_df, path, compression_level=22):
        self.writer.parallel_write_chromdata_to_parquet(
            data_df,
            path,
            chrom_names=[x for x in self.gp.chrom_sizes.keys()],
            compression_level=compression_level,
        )



import numpy as np
from scipy.signal import convolve
import pywt


def convolve1d_fast(arr, ker, mode='wrap', origin=0):
    pad_width = len(ker) // 2
    padded_arr = np.pad(arr, pad_width, mode=mode)
    conv_result = convolve(padded_arr, ker, mode='valid')
    return np.roll(conv_result[:len(arr)], origin)


def convolve_s(g_t, v_j, j):
    """
    (j-1)th level synthesis from w_j
    """
    dtype = v_j.dtype
    g_ker = np.zeros(len(g_t) * 2**(j - 1), dtype=dtype)

    for i, g in enumerate(g_t):
        g_ker[i * 2**(j - 1)] = g

    return convolve1d_fast(
        v_j,
        np.flip(g_ker),
        mode='wrap',
        origin=(len(g_ker) - 1) // 2
    )


def init_g_t(filters):
    wavelet = pywt.Wavelet(filters)
    g = wavelet.dec_lo
    return np.array(g, dtype=np.float32) / np.sqrt(2)


def imodwt(v_j, filters, level):
    ''' inverse modwt '''
    g_t = init_g_t(filters)
    for jp in range(level):
        j = level - jp - 1
        v_j = convolve_s(g_t, v_j, j + 1)
    return v_j


def convolve_d(h_t, v_j_1, j):
    '''
    jth level decomposition
    h_t: \tilde{h} = h / sqrt(2)
    v_j_1: v_{j-1}, the (j-1)th scale coefficients
    return: w_j (or v_j)
    '''
    dtype = v_j_1.dtype
    ker = np.zeros(len(h_t) * 2**(j - 1), dtype=dtype)
    for i, h in enumerate(h_t):
        ker[i * 2**(j - 1)] = h

    return convolve1d_fast(v_j_1, ker, mode='wrap', origin=-len(ker) // 2)


def modwt(x, filters, level):
    '''
    filters: 'db1', 'db2', 'haar', ...
    return: see matlab
    '''
    g_t = init_g_t(filters)
    v_j_1 = x
    for j in range(level):
        v_j_1 = convolve_d(g_t, v_j_1, j + 1)
    return v_j_1


def modwt_smooth(x, filters, level):
    w = modwt(x, filters, level) # last level approximation
    return imodwt(w, filters, level)



"""
This script is used to produce q-values corrected for the set of samples
"""
from hotspot3.io.logging import WithLogger
from hotspot3.io.readers import GenomeReader
from hotspot3.io.writers import GenomeWriter
from hotspot3.io import parallel_write_partitioned_parquet, read_partioned_parquet

from hotspot3.scoring import find_potentialy_significant_pvals, logfdr_from_logpvals
from hotspot3.helpers.models import SampleFDRdata, MultiSampleFDRData, FDRData
from hotspot3.config import ProcessorConfig

import numpy as np
import pandas as pd
import shutil
from logging import Logger
import gc
from concurrent.futures import ProcessPoolExecutor, as_completed, ThreadPoolExecutor
from typing import List
from genome_tools import GenomicInterval
import os
from tqdm import tqdm
import multiprocessing


class FDRCorrection(WithLogger):

    def __init__(self, name, config: ProcessorConfig=None, logger: Logger=None, *, chrom_sizes: dict=None):
        super().__init__(config, logger, name)
        self.chrom_sizes = chrom_sizes
        self.reader = self.copy_with_params(GenomeReader)
        self.writer = self.copy_with_params(GenomeWriter)

    def compute_fdr(self, fdr_data: FDRData) -> np.ndarray:
        potentially_signif_pvals = fdr_data.potentially_signif_pvals
        potentially_signif_pvals *= -np.log(10)

        logfdr = logfdr_from_logpvals(
            potentially_signif_pvals,
            method=self.config.fdr_method,
            m=fdr_data.n_tests
        ).astype(np.float16)
        logfdr /= -np.log(10)
        
        return logfdr
    
    def fdr_correct_pvals(self):
        raise NotImplementedError("This method should be implemented in a subclass")
    
    def extract_data_for_sample(self):
        raise NotImplementedError("This method should be implemented in a subclass")


class SampleFDRCorrection(FDRCorrection):

    def fdr_correct_pvals(self, pvals_path, max_fdr, save_path):

        mask_path = f"{save_path}.mask"
        self.writer.sanitize_path(mask_path)
        self.logger.debug(f"{self.name}: Extracting raw P-values")
        fdr_data, mask = self.extract_data_for_sample(pvals_path, max_fdr, mask_path, return_mask=True)

        self.logger.debug(f"{self.name}: Computing FDR")
        result = self.compute_fdr(fdr_data)
        result = self.cast_to_original_shape(result, mask)

        self.writer.sanitize_path(save_path)
        self.logger.debug(f"{self.name}: Saving FDRs to parquet")
        self.write_fdr_partitioned_by_sample_and_chrom(
            fdr_data.chrom_pos_mapping,
            result,
            save_path,
            cpus=self.config.cpus
        )
    
    def read_fdrs_for_chrom(self, fdr_path, chrom):
        return read_partioned_parquet(
            fdr_path,
            partition_cols=['chrom', 'sample_id'],
            partition_vals=[chrom, self.name],
            columns=['log10_fdr']
        )['log10_fdr'].values


    def extract_data_for_sample(self, pvals_path, fdr, save_path, all_ids=None, return_mask=False):
        if all_ids is None:
            all_ids = [self.name]
        log_pvals = self.reader.read_pval_from_parquet(pvals_path)
        mask, n_tests = find_potentialy_significant_pvals(log_pvals, fdr)
        log_pvals = log_pvals[mask]

        chrom_pos_mapping = self.reader.read_chrom_pos_mapping(
            pvals_path,
            chrom_sizes=self.chrom_sizes,
        )
        data = SampleFDRdata(log_pvals, n_tests, self.name, chrom_pos_mapping)
        if return_mask:
            return data, mask
        self.logger.info(f"{self.name}: Saving mask")
        
        self.write_mask_data_to_np(mask, save_path)
        return data
    
    def write_mask_data_to_np(self, mask, save_path):
        os.makedirs(save_path, exist_ok=True)
        np.save(os.path.join(save_path, f"{self.name}.npy"), mask)

    def extract_mask_for_sample(self, mask_path) -> np.ndarray:
        return np.load(os.path.join(mask_path, f"{self.name}.npy"))
    
    def cast_to_original_shape(self, result: np.ndarray, mask) -> np.ndarray:
        if isinstance(mask, str):
            mask = self.extract_mask_for_sample(mask)
        res = np.full(mask.shape, np.nan, dtype=result.dtype)
        res[mask] = result
        return res
    
    def write_fdr_partitioned_by_sample_and_chrom(
            self,
            chrom_pos_mapping: List[GenomicInterval],
            fdrs: np.ndarray,
            save_path,
            all_ids=None,
            cpus=None
        ):
        if all_ids is None:
            all_ids = [self.name]
        
        if cpus is None:
            cpus = self.config.cpus

        if cpus == 1:
            for gi in chrom_pos_mapping:
                chrom_df = self.df_from_fdrs(fdrs, gi, all_ids)
                parallel_write_partitioned_parquet(
                    chrom_df,
                    field_names=[gi.chrom, self.name],
                    partition_cols=['chrom', 'sample_id'],
                    path=save_path,
                    tmp_dir=self.config.tmp_dir
                )
        else:
            params = [
                (
                    self.df_from_fdrs(fdrs, gi, all_ids),
                    (gi.chrom, self.name),
                    ['chrom', 'sample_id'],
                    save_path,
                    self.config.tmp_dir
                )
                for gi in chrom_pos_mapping
            ]
            params = [list(x) for x in zip(*params)]
            with ProcessPoolExecutor(cpus) as executor:
                [x for x in executor.map(parallel_write_partitioned_parquet, *params)]
        
    def df_from_fdrs(self, fdrs, genomic_interval: GenomicInterval, all_ids: List[str]) -> pd.DataFrame:
        sample_id_codes = np.full(len(genomic_interval), all_ids.index(self.name), dtype=np.int16)
        
        chrom_categories = list(self.chrom_sizes.keys())
        chrom_codes = np.full(
            len(genomic_interval),
            chrom_categories.index(genomic_interval.chrom),
            dtype=np.int16
        )
        return pd.DataFrame({
            'log10_fdr': fdrs[genomic_interval.start:genomic_interval.end],
            'chrom': pd.Categorical.from_codes(chrom_codes, categories=chrom_categories),
            'sample_id': pd.Categorical.from_codes(sample_id_codes, categories=all_ids)
        })


class MultiSampleFDRCorrection(FDRCorrection):
    def fdr_correct_pvals(self, paths, fdr_cutoff, save_path):
        if fdr_cutoff > 0.05:
            self.logger.warning(f"Requested FDR is higher than 0.05: {fdr_cutoff}. The code is not optimized for high FDRs with high number of samples")

        mask_path = f"{save_path}.mask"
        self.writer.sanitize_path(mask_path)
        self.logger.info(f"Extracting data for FDR correction using {self.config.cpus} CPUs")
        self.logger.info(f"Extracting data for {len(paths)} samples")
        all_samples_fdr_data = self.extract_data_for_sample(paths, fdr_cutoff, mask_path)
        
        logfdr = self.compute_fdr(all_samples_fdr_data)

        all_samples_fdr_data.potentially_signif_pvals = None
        gc.collect()
        
        self.writer.sanitize_path(save_path)
        self.write_fdr_partitioned_by_sample_and_chrom(
            all_samples_fdr_data,
            logfdr,
            mask_path,
            save_path
        )
        shutil.rmtree(mask_path) # cleanup

    
    def process_sample(self, sample_id, pvals_path, fdr, save_path):
        fdr_correction_data = SampleFDRCorrection(
            name=sample_id,
            chrom_sizes=self.chrom_sizes,
            logger=self.logger,
            config=self.config
        ).extract_data_for_sample(
            pvals_path,
            fdr,
            all_ids=self.name,
            save_path=save_path,
            return_mask=False
        )
        return fdr_correction_data

    def extract_data_for_sample(self, paths: dict, fdr, save_path):
        self.writer.sanitize_path(save_path)

        all_args = [(sample_id, paths[sample_id], fdr, save_path) for sample_id in self.name]
        all_args = [list(x) for x in zip(*all_args)]

        if self.config.cpus > 1:
            with ThreadPoolExecutor(
                max_workers=self.config.cpus,
            ) as executor:
                try:
                    results_list = {}
                    futures = {
                        executor.submit(self.process_sample, *args): sample_id
                        for sample_id, args in zip(self.name, zip(*all_args))
                    }

                    with tqdm(total=len(futures), desc="Processing Samples") as pbar:
                        for future in as_completed(futures):
                            sample_id = futures[future]
                            try:
                                result = future.result()
                                self.logger.debug(f"Data extracted for {sample_id}")
                                results_list[sample_id] = result
                            except Exception as e:
                                self.logger.error(f"Error processing {sample_id}: {e}")
                                raise e
                            finally:
                                pbar.update(1)
                except Exception as e:
                    self.logger.critical(f"Exception occured, gracefully shutting down executor...")
                    self.logger.critical(e)
                    raise e
  
        else:
            results_list = {
                args[0]: self.process_sample(*args) 
                for args in zip(*all_args)
            }

        results = []
        chrom_pos_mappings = {}
        n_tests = 0
        sample_id_correspondance = pd.DataFrame(
            {'start_index': pd.NA, 'end_index': pd.NA},
            index=self.name,
        )
        current_index = 0
        self.logger.info('Concatenating data for all samples')
        for sample_id in self.name:
            fdr_correction_data = results_list[sample_id]
            potentially_significant_pvals = fdr_correction_data.potentially_signif_pvals
            n_tests += fdr_correction_data.n_tests

            sample_id_correspondance.loc[sample_id, 'start_index'] = current_index
            current_index += len(potentially_significant_pvals)
            sample_id_correspondance.loc[sample_id, 'end_index'] = current_index

            chrom_pos_mappings[sample_id] = fdr_correction_data.chrom_pos_mapping

            results.append(potentially_significant_pvals)

        self.logger.debug(f"Data extracted for {len(paths)} samples")
     
        potentially_significant_pvals = np.concatenate(results)

        sample_id_correspondance = sample_id_correspondance.astype(np.int64)
        return MultiSampleFDRData(
            potentially_significant_pvals,
            n_tests,
            sample_id_correspondance,
            chrom_pos_mappings
        )

    def write_fdr_partitioned_by_sample_and_chrom(
            self,
            all_samples_fdr_data: MultiSampleFDRData,
            logfdr: np.ndarray,
            mask_path,
            save_path
        ):
        sample_id_correspondance = all_samples_fdr_data.sample_id_correspondance
        for sample_id, row in sample_id_correspondance.iterrows():
            self.logger.debug(f"Writing FDR for {sample_id}")
            sample_correction = self.copy_with_params(
                SampleFDRCorrection,
                name=sample_id,
                chrom_sizes=self.chrom_sizes
            )
            sample_fdrs = logfdr[row['start_index']:row['end_index']]
            sample_fdrs = sample_correction.cast_to_original_shape(sample_fdrs, mask_path)

            sample_correction.write_fdr_partitioned_by_sample_and_chrom(
                chrom_pos_mapping=all_samples_fdr_data.chrom_pos_mappings[sample_id],
                fdrs=sample_fdrs,
                save_path=save_path,
                all_ids=self.name
            )



import numpy as np
import numpy.ma as ma
import dataclasses

from hotspot3.background_fit.fit import StridedBackgroundFit as StridedFit
from hotspot3.helpers.models import FitResults, WindowedFitResults, DataForFit
from hotspot3.helpers.utils import wrap_masked


class StridedBackgroundFit(StridedFit):

    ## Currently deprecated methods ##
    @wrap_masked
    def fit(self, agg_cutcounts: np.ndarray, fallback_fit_results: FitResults=None):
        data_for_fit = self.prepare_data_for_fit(agg_cutcounts)

        best_fit_results = self.find_best_fits(
            data_for_fit,
            fallback_fit_results=fallback_fit_results,
        )

        if fallback_fit_results is not None:
            best_fit_results.fit_threshold = np.where(
                best_fit_results.rmsea <= self.config.rmsea_tr,
                best_fit_results.fit_threshold , 
                self.fallback_tr(data_for_fit, best_fit_results, fallback_fit_results)
            )

        return self.cast_to_original_shape(best_fit_results, agg_cutcounts.shape)

    def cast_to_original_shape(
            self,
            fit_results: FitResults,
            original_shape: tuple
        ) -> FitResults:
        subsampled_indices = self.get_subsampled_indices(original_shape)

        for data_field in dataclasses.fields(fit_results):
            field = data_field.name
            values = getattr(fit_results, field)
            if values is not None and not (np.ndim(values) == 0 and np.isnan(values)):
                upcasted = self.upcast(original_shape, subsampled_indices, values)
                setattr(fit_results, field, upcasted)
        return fit_results
    
    def prepare_data_for_fit(self, agg_cutcounts: ma.MaskedArray):
        strided_agg_cutcounts = self.get_strided_agg_cutcounts(agg_cutcounts)
        max_counts_with_flanks = self.get_max_count_with_flanks(agg_cutcounts)
        strided_max_counts_with_flanks = self.get_strided_agg_cutcounts(max_counts_with_flanks)
        bin_edges, n_signal_bins = self.get_all_bins(strided_agg_cutcounts)
        value_counts = self.value_counts_per_bin(strided_agg_cutcounts, bin_edges)
        self.merge_signal_bins_for_rmsea_inplace(value_counts, bin_edges, n_signal_bins)
        return DataForFit(bin_edges, value_counts, n_signal_bins, strided_agg_cutcounts, strided_max_counts_with_flanks)


    def find_best_fits(
            self,
            data_for_fit: DataForFit,
            fallback_fit_results: FitResults=None
        ) -> FitResults:

        best_fit_results = FitResults(
            None,
            None,
            rmsea=np.full(data_for_fit.agg_cutcounts.shape[1], np.inf, dtype=np.float32),
            fit_threshold=np.asarray(data_for_fit.bin_edges[-1], dtype=np.float32),
        )
        remaing_fits_mask = np.ones_like(best_fit_results.fit_threshold, dtype=bool)

        for i in np.arange(0, data_for_fit.n_signal_bins, 1)[::-1]:
            if remaing_fits_mask.sum() == 0:
                break

            # 1. Prepare data for the current threshold
            current_index = data_for_fit.bin_edges.shape[0] - i - 1
            changing_indices = self._get_changing_indices(current_index, data_for_fit, remaing_fits_mask)

            data_for_current_threshold, assumed_signal_mask = self.get_data_for_current_threshold(
                data_for_fit,
                current_index,
                changing_indices
            )
            
            # 2. Fit for the current threshold and evaluate the fit
            bin_fit_results = self.fit_for_bin(
                data_for_current_threshold.agg_cutcounts,
                where=~assumed_signal_mask,
                fallback_fit_results=fallback_fit_results
            )
            enough_bg_mask = bin_fit_results.enough_bg_mask


            rmsea = self.evaluate_fit_for_bin(
                data_for_current_threshold,
                bin_fit_results,
                fallback_fit_results,
            )
            bin_fit_results = FitResults(
                bin_fit_results.p,
                bin_fit_results.r,
                rmsea=rmsea,
                fit_threshold=data_for_current_threshold.bin_edges[-1]
            )

            # 3. Update the best fit results
            self.update_remaining_fits(
                best_fit_results,
                remaing_fits_mask,
                bin_fit_results,
                enough_bg_mask,
                changing_indices
            )

            idx = data_for_fit.n_signal_bins - i
            if idx % (max(data_for_fit.n_signal_bins, 5) // 5) == 0:
                self.logger.debug(f"{self.name} (window={self.config.bg_window}): Identifying signal proportion (step {idx}/{data_for_fit.n_signal_bins})")
        
        best_fit_results.fit_quantile = self.get_bg_quantile_from_tr(
            data_for_fit.agg_cutcounts,
            best_fit_results.fit_threshold
        )
        return best_fit_results
    
    def get_data_for_current_threshold(
            self,
            data_for_fit: DataForFit,
            current_index,
            changing_indices
    ):
        right_bin_index = current_index + 1

        current_agg_cutcounts = data_for_fit.agg_cutcounts[:, changing_indices]
        current_bin_edges = data_for_fit.bin_edges[:right_bin_index, changing_indices]
        currect_counts_with_flanks = data_for_fit.max_counts_with_flanks[:, changing_indices]
        
        assumed_signal_mask = currect_counts_with_flanks >= current_bin_edges[-1]

        # TODO: don't update the value counts from first iteration
        current_value_counts = self.value_counts_per_bin(
            current_agg_cutcounts,
            current_bin_edges,
            where=~assumed_signal_mask
        )

        return DataForFit(
            current_bin_edges,
            current_value_counts,
            data_for_fit.n_signal_bins,
            current_agg_cutcounts,
            currect_counts_with_flanks
        ), assumed_signal_mask


    def _get_changing_indices(self, current_index, data_for_fit: DataForFit, remaing_fits_mask: np.ndarray):
        fit_will_change = data_for_fit.value_counts[current_index - 1][remaing_fits_mask] != 0 # shape of remaining_fits
        changing_indices = np.where(remaing_fits_mask)[0][fit_will_change]
        return changing_indices
    
    def fit_for_bin(
            self,
            strided_agg_cutcounts: np.ndarray,
            where: np.ndarray=None,
            fallback_fit_results: FitResults=None
        ):
        window = min(self.config.bg_window, strided_agg_cutcounts.shape[0])
        min_count = self.get_min_count(window / self.sampling_step)
        enough_bg_mask = np.sum(
            ~np.isnan(strided_agg_cutcounts, where=where),
            axis=0,
            where=where
        ) > min_count

        mean = np.full(strided_agg_cutcounts.shape[1], np.nan, dtype=np.float32)
        var = np.full(strided_agg_cutcounts.shape[1], np.nan, dtype=np.float32)
        mean[enough_bg_mask], var[enough_bg_mask] = self.get_mean_and_var(
            strided_agg_cutcounts[:, enough_bg_mask],
            where=where[:, enough_bg_mask]
        )

        if fallback_fit_results is not None:
            r = np.full_like(mean, fallback_fit_results.r, dtype=np.float32)
        else:
            r = self.r_from_mean_and_var(mean, var)

        p = self.p_from_mean_and_r(mean, r)

        return WindowedFitResults(p, r, enough_bg_mask=enough_bg_mask)
    
    def evaluate_fit_for_bin(
            self,
            data_for_fit: DataForFit,
            fit_results: WindowedFitResults,
            fallback_fit_results: FitResults,
        ):
        n_params = 1 if fallback_fit_results is not None else 2
        rmsea = np.full_like(fit_results.p, np.nan)

        rmsea[fit_results.enough_bg_mask] = self.calc_rmsea_all_windows(
            fit_results.p[fit_results.enough_bg_mask],
            fit_results.r[fit_results.enough_bg_mask],
            n_params=n_params,
            bin_edges=data_for_fit.bin_edges[:, fit_results.enough_bg_mask],
            value_counts_per_bin=data_for_fit.value_counts[:, fit_results.enough_bg_mask]
        )
        return rmsea
    
    def update_remaining_fits(
            self,
            best_fit_results: FitResults,
            remaing_fits_mask: np.ndarray,
            bin_fit_results: FitResults,
            enough_bg_mask: np.ndarray,
            changing_indices: np.ndarray,
        ):
        remaining_fits = enough_bg_mask & ~(bin_fit_results.rmsea <= self.config.rmsea_tr)
        remaing_fits_mask[changing_indices] = remaining_fits

        better_fit = bin_fit_results.rmsea <= best_fit_results.rmsea[changing_indices]
        self.update_best_fit(
            best_fit_results, bin_fit_results,
            changing_indices,
            where=better_fit
        )
        self.update_best_fit(
            best_fit_results,
            FitResults(),
            changing_indices,
            where=~enough_bg_mask
        )
    
    def update_best_fit(
            self,
            best_fit, current_fit, changing_indices, where
        ): 
        for field in ['rmsea', 'fit_threshold']:
            current = getattr(current_fit, field)
            best = getattr(best_fit, field)[changing_indices]
            best = np.where(
                where,
                current,
                best
            )

    def fallback_tr(
            self,
            data_for_fit: DataForFit,
            fit_results: FitResults,
            fallback_fit_results: FitResults
        ):
        global_quantile_tr = self.quantile_ignore_all_na(
            data_for_fit.agg_cutcounts,
            fallback_fit_results.fit_quantile
        )
        return np.where(
            global_quantile_tr < fallback_fit_results.fit_threshold,
            fallback_fit_results.fit_threshold,
            np.where(
                fit_results.fit_threshold < global_quantile_tr,
                fit_results.fit_threshold,
                self.quantile_ignore_all_na(data_for_fit.agg_cutcounts, self.config.min_background_prop) 
            )
        )


import numpy as np
import numpy.ma as ma
from typing import List

from genome_tools import GenomicInterval

from hotspot3.helpers.models import FitResults, NotEnoughDataForContig
from hotspot3.io.logging import WithLoggerAndInterval
from hotspot3.connectors.bottleneck import BottleneckWrapper

from babachi.segmentation import GenomeSegmentator
from babachi.models import GenomeSNPsHandler, ChromosomeSNPsHandler


class BabachiWrapper(WithLoggerAndInterval):

    def craft_snps_collection(
            self,
            agg_cutcounts: ma.MaskedArray,
            per_window_trs: np.ndarray,
            fit_results: FitResults,
        ):
        step = self.config.babachi_segmentation_step

        bn_wrapper = self.copy_with_params(BottleneckWrapper)
        assumed_signal_mask = bn_wrapper.get_signal_mask_for_tr(agg_cutcounts, per_window_trs)
        background = agg_cutcounts.filled(np.nan)[::step]
        background[assumed_signal_mask[::step]] = np.nan
        starts = np.arange(0, len(background), dtype=np.uint32) * step
        valid_counts = ~np.isnan(background)
        
        chrom_handler = ChromosomeSNPsHandler(
            self.genomic_interval.chrom,
            positions=starts[valid_counts], 
            read_counts=np.stack(
                [
                    np.full(background.shape[0], fit_results.r, dtype=np.float32),
                    background
                ]
            ).T[valid_counts, :]
        )
        return GenomeSNPsHandler(chrom_handler)

    def run_segmentation(
            self,
            agg_cutcounts: ma.MaskedArray,
            per_window_trs: np.ndarray,
            fit_results: FitResults
        ):
        bads, chrom_bad = self.get_bads(fit_results)
        snps_collection = self.craft_snps_collection(
            agg_cutcounts,
            per_window_trs,
            fit_results
        )
        chrom_sizes = {self.genomic_interval.chrom: len(self.genomic_interval)}
        bad_segments = self.run_babachi(snps_collection, chrom_sizes, bads)
        if len(bad_segments) == 0:
            raise NotEnoughDataForContig
        return [
            GenomicInterval(x.chr, x.start, x.end, BAD=x.BAD / chrom_bad) for x in bad_segments
        ]
    
    def get_bads(self, fit_results: FitResults, max_bad=10):
        chrom_bad = fit_results.p / (1 - fit_results.p)
        mult = np.arange(1, max_bad + 0.1, 1)
        return [*(mult * chrom_bad), *(1 / mult[1:] * chrom_bad)], chrom_bad
    
    def run_babachi(self, snps_collection, chrom_sizes, bads):
        gs = GenomeSegmentator(
            snps_collection=snps_collection,
            chrom_sizes=chrom_sizes,
            jobs=1,
            logger_level=self.config.logger_level,
            segmentation_mode='binomial',
            states=bads,
            logger=self.logger,
            allele_reads_tr=0,
            b_penalty=self.config.babachi_boundary_penalty,
            min_seg_bp=self.config.babachi_min_segment_size,
            min_seg_snps=0,
            subchr_filter=self.config.babachi_min_segment_size // self.config.babachi_segmentation_step
        )
        return gs.estimate_BAD()

    def annotate_with_segments(self, shape, bad_segments: List[GenomicInterval]):
        babachi_result = np.zeros(shape, dtype=np.float16)
        for segment in bad_segments:
            babachi_result[segment.start:segment.end] = segment.BAD
        return babachi_result



"""
Folder that contains connectors to different libraries:
- bottleneck: bottleneck library for fast computation of moving statistics
- babachi: babachi library for signal segmentation
- bam2bed: bam2bed library for conversion of bam files to bed files
"""


import numpy as np


def rolling_view_with_nan_padding(arr, points_in_window=501, interpolation_step=1000):
    """
    Creates a 2D array where each row is a shifted view of the original array with NaN padding.
    Only every 'subsample_step' row is taken to reduce memory usage.

    Parameters:
        arr (np.ndarray): Input array of shape (n,)
        window_size (int): The total size of the shift (default is 501 for shifts from -250 to +250)
        subsample_step (int): Step size for subsampling (default is 1000)

    Returns:
        np.ndarray: A subsampled array with NaN padding, of shape (n // subsample_step, window_size)
    """
    n = arr.shape[0]
    assert points_in_window % 2 == 1, "Window size must be odd to have a center shift of 0."
    
    pad_width = (points_in_window - 1) // 2
    padded_arr = np.pad(arr, pad_width, mode='constant', constant_values=np.nan)
    subsample_count = (n + interpolation_step - 1) // interpolation_step
    shape = (subsample_count, points_in_window)
    strides = (padded_arr.strides[0] * interpolation_step, padded_arr.strides[0])
    subsampled_view = np.lib.stride_tricks.as_strided(padded_arr, shape=shape, strides=strides)
 
    return subsampled_view.T


def calc_g_sq(obs, exp):
    valid = (exp != 0) & (obs != 0)
    with np.errstate(over='ignore'):
        ratio = np.divide(obs, exp, out=np.zeros_like(obs), where=valid)
    ratio = np.where(valid, ratio, 1)
    return obs * np.log(ratio) * 2


def calc_chisq(obs, exp):
    with np.errstate(over='ignore'):
        return np.where((exp != 0) & (obs != 0), (obs - exp) ** 2 / exp, 0)


def calc_rmsea(obs, exp, N, df, min_df=1, stat='G_sq', where=None):
    assert stat in ('G_sq', 'chi_sq'), "Only G_sq and chi_sq statistics are supported"
    if where is None:
        where = np.ones_like(obs, dtype=bool)
    if stat == 'G_sq':
        G_sq = calc_g_sq(obs, exp)
    else:
        G_sq = calc_chisq(obs, exp)

    G_sq = np.sum(G_sq, axis=0, where=where)
    G_sq = np.divide(G_sq, df, out=np.zeros_like(G_sq), where=df>=min_df)
    rmsea = np.sqrt(np.maximum(G_sq - 1, 0) / (N - 1))
    rmsea = np.where(df >= min_df, rmsea, np.inf)
    return rmsea



import numpy as np
import scipy.stats as st
import gc
from scipy.special import logsumexp, betainc, hyp2f1, betaln


# Calculate p-values and FDR
def logpval_for_dtype(x: np.ndarray, r: np.ndarray, p: np.ndarray, dtype=None, calc_type="betainc") -> np.ndarray:
    """
    Implementation of log(pval) with high precision
    """
    mask = x > 0
    x = np.round(np.asarray(x, dtype=dtype))
    r = np.broadcast_to(np.asarray(r, dtype=dtype), x.shape)[mask]
    p = np.broadcast_to(np.asarray(p, dtype=dtype), x.shape)[mask]
    x = x[mask]
    
    result = np.zeros(mask.shape, dtype=dtype)
    if calc_type == 'nbinom':
        p_vals = st.nbinom.logsf(x - 1, r, 1 - p)
    elif calc_type == 'betainc':
        p_vals = logpval_for_dtype_betainc(x, r, p)
    elif calc_type == 'hyp2f1':
        p_vals = logpval_for_dtype_hyp2f(x, r, p)
    else:
        raise ValueError(f"Unknown p-value calculation type: {calc_type}")
    result[mask] = p_vals
    return result


def logpval_for_dtype_betainc(x: np.ndarray, r: np.ndarray, p: np.ndarray) -> np.ndarray:
    with np.errstate(divide='ignore', invalid='ignore'):
        return np.log(betainc(x, r, p, dtype=r.dtype))


def logpval_for_dtype_hyp2f(x: np.ndarray, r: np.ndarray, p: np.ndarray) -> np.ndarray:
    with np.errstate(divide='ignore', invalid='ignore'):
        return (
            x * np.log(p) 
            + r * np.log(1 - p) 
            + np.log(hyp2f1(x + r, 1, x + 1, p, dtype=r.dtype))
            - np.log(x)
            - betaln(x, r, dtype=r.dtype)
        )

def find_potentialy_significant_pvals(log_pval: np.ndarray, max_fdr: float):
    non_nan_pvals = ~np.isnan(log_pval)
    number_of_tests = np.sum(non_nan_pvals)
    return non_nan_pvals & (log_pval >= -np.log10(max_fdr)), number_of_tests

def fast_logfdr_below_threshold(log_pval: np.ndarray, max_fdr: float, fdr_method: str):
    result = np.full_like(log_pval, np.nan)
    mask, m = find_potentialy_significant_pvals(log_pval, max_fdr)

    # Convert to natural log
    log_pval = log_pval[mask]
    log_pval *= -np.log(10)

    result[mask] = logfdr_from_logpvals(log_pval, method=fdr_method, m=m)
    result /= -np.log(10)
    return result


def logfdr_from_logpvals(log_pvals, *, method='bh', dtype=np.float32, m=None) -> np.ndarray:
    """
    Reimplementation of scipy.stats.false_discovery_control to work with log-transformed p-values.
    Accepts log-transformed p-values and returns log-transformed FDR values.
    NOTE: Not log10-transformed and not negated!

    Parameters:
        - log_pvals: 1D array of log-transformed p-values. Most of them will be negative.
        - fdr_method: Method to use for FDR calculation. 'bh' = Benjamini-Hochberg, 'by' = Benjamini-Yekutieli.
    
    Returns:
        - log_fdr: 1D array of log-transformed FDR values. 
    """
    assert method in ['bh', 'by'], "Only 'bh' and 'by' methods are supported."
    log_pvals = np.asarray(log_pvals, dtype=dtype) # no further input checking yet
    if log_pvals.ndim != 1:
        raise NotImplementedError("Only 1D arrays are supported.")

    if m is None:
        m = log_pvals.shape[0]
    order = np.argsort(log_pvals) # can save argsort as uint32
    
    log_pvals = log_pvals[order]

    log_i = np.log(np.arange(1, log_pvals.shape[0] + 1, dtype=dtype))
    log_pvals += np.log(m).astype(dtype) - log_i  # p_adj = p * m / i => log10(p_adj) = log10(p) + log10(m) - log10(i)
    if method == 'by':
        log_pvals += logsumexp(-log_i)
    del log_i
    gc.collect()
    np.minimum.accumulate(log_pvals[::-1], out=log_pvals[::-1])
    log_pvals_copy = log_pvals.copy() # since array is float32, copying is more memory effient than argsort int64
    log_pvals[order] = log_pvals_copy
    return np.clip(log_pvals, a_min=None, a_max=0)


def fix_inf_pvals(neglog_pvals):
    infs = np.isinf(neglog_pvals)
    n_infs = np.sum(infs) 
    if n_infs > 0:
        neglog_pvals[infs] = 300
    return neglog_pvals




import dataclasses
import pandas as pd
import numpy as np
from typing import List, Dict, AnyStr
from genome_tools import GenomicInterval


@dataclasses.dataclass
class FitResults:
    p: np.ndarray = np.nan
    r: np.ndarray = np.nan
    rmsea: np.ndarray = np.nan
    fit_quantile: np.ndarray = np.nan
    fit_threshold: np.ndarray = np.nan
    n_signal: int = np.nan # Number of signal bases
    n_total: int = np.nan # Total number of mappable bases
    signal_tags: int = np.nan # Number of tags at signal bases
    total_tags: int = np.nan # Number of tags total


@dataclasses.dataclass
class WindowedFitResults:
    p: np.ndarray
    r: np.ndarray
    enough_bg_mask: np.ndarray


@dataclasses.dataclass
class DataForFit:
    bin_edges: np.ndarray
    value_counts: np.ndarray
    n_signal_bins: int
    agg_cutcounts: np.ndarray
    max_counts_with_flanks: np.ndarray

@dataclasses.dataclass
class SPOTEstimationData:
    total_tags: np.ndarray
    total_tags_background: np.ndarray
    segment_spot_scores: np.ndarray
    total_bases: np.ndarray
    valid_scores: np.ndarray

@dataclasses.dataclass
class SPOTEstimationResults:
    spot_score: float
    spot_score_std: float
    slope: float
    intercept: float
    inliers_mask: np.ndarray
    outlier_distance: np.ndarray
    min_bg_tags_fraction: np.ndarray


@dataclasses.dataclass
class FDRData:
    potentially_signif_pvals: np.ndarray
    n_tests: int

@dataclasses.dataclass
class SampleFDRdata(FDRData):
    sample_id: str
    chrom_pos_mapping: List[GenomicInterval]


@dataclasses.dataclass
class MultiSampleFDRData(FDRData):
    sample_id_correspondance: pd.DataFrame
    chrom_pos_mappings: Dict[AnyStr, List[GenomicInterval]]


@dataclasses.dataclass
class ProcessorOutputData:
    """
    Dataclass for storing the output of ChromosomeProcessor and GenomeProcessor methods.
    """
    id: str
    data_df: pd.DataFrame

class NotEnoughDataForContig(Exception):
    """Exception raised when a required contig is not present."""

    def __init__(self, message="No contig is present in the provided data."):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"{self.__class__.__name__}: {self.message}"






import logging
import sys
from typing import Type, TypeVar

from genome_tools import GenomicInterval

from hotspot3.config import ProcessorConfig


T = TypeVar('T', bound='WithLogger')

class WithLogger:
    def __init__(self, config=None, logger=None, name=None):
        if config is None:
            config = ProcessorConfig()
        self.config = config

        self._logger = logger if logger is not None else setup_logger(level=self.config.logger_level)

        if name is None:
            name = self.__class__.__name__
        self.name = name
    
    @property
    def logger(self) -> logging.Logger:
        return self._logger

    @logger.setter
    def logger(self, logger: logging.Logger):
        # Setter allows updating the logger
        self._logger = logger
    
    def __getstate__(self):
        state = self.__dict__
        if 'logger' in state:
            del state['logger']
        return state

    def __setstate__(self, state: dict):
        for name, value in state.items():
            setattr(self, name, value)


    def copy_with_params(self, cls: Type[T], **kwargs) -> T:
        """
        Creates a new instance of the specified class `cls` with the same initialization parameters.
        """
        assert issubclass(cls, WithLogger), "cls should be a subclass of WithLogger"
        data = dict(
            config=self.config,
            logger=self.logger,
            name=self.name
        )
        data.update(kwargs)
        return cls(**data)


class WithLoggerAndInterval(WithLogger):
    def __init__(self, genomic_interval: GenomicInterval, config: ProcessorConfig=None, logger=None):
        super().__init__(logger=logger, config=config, name=genomic_interval.to_ucsc())
        self.genomic_interval = genomic_interval
    
    def copy_with_params(self, cls: Type[T], **kwargs) -> T:
        """
        Creates a new instance of the specified class `cls` with the same initialization parameters.
        """
        if issubclass(cls, WithLoggerAndInterval):
            class_fields = dict(
                genomic_interval=self.genomic_interval,
                config=self.config,
                logger=self.logger
            )
        elif issubclass(cls, WithLogger):
            class_fields = dict(
                config=self.config,
                logger=self.logger,
                name=self.genomic_interval.to_ucsc()
            )
        else:
            raise ValueError(f"cls should be a subclass of WithLogger or WithLoggerAndInterval, got {cls}")
        class_fields.update(kwargs)
        return cls(**class_fields)


def setup_logger(name='hotspot3', level=None, outstream=None) -> logging.Logger:
    if outstream is None:
        outstream = sys.stdout

    logger = logging.getLogger(name)
    if level is not None:
        logger.setLevel(level)
        if not logger.handlers:
            handler = logging.StreamHandler(outstream)
            handler.setLevel(level)
            formatter = logging.Formatter('%(asctime)s  %(levelname)s  %(message)s')
            handler.setFormatter(formatter)
            logger.addHandler(handler)
    return logger



import matplotlib.pyplot as plt


def get_bb_color(fdr, mode='peaks'):
    thresholds = [0.001, 0.01, 0.05, 0.1]
    cmap_tab20c = plt.get_cmap("tab20c")

    if mode == 'peaks':
        indices = [8, 9, 10, 11]
    elif mode == 'hotspots':
        indices = [12, 13, 14, 15]
    else:
        raise ValueError("Mode must be either 'peaks' or 'hotspots'.")

    colors = [
        *[tuple(int(c * 255) for c in cmap_tab20c(i)[:3]) for i in indices], (192, 192, 192)
    ]
    
    assert len(colors) == len(thresholds) + 1, "Number of colors must be one more than the number of thresholds."
    for threshold, color in zip(thresholds, colors):
        if fdr <= threshold:
            return "{},{},{}".format(*color)
    return "{},{},{}".format(*colors[-1])


import dataclasses
import logging


@dataclasses.dataclass
class ProcessorConfig:
    """
    Parameters:
        - window: Bandwidth for signal smoothing.
        - min_mappable: Minimum number of mappable bases for a window to be tested.
    
        - bg_window: Window size for aggregating background cutcounts.
        - min_mappable_bg: Minimum number of mappable bases for a window to be considered in background.

        - density_step: Step size for extracting density.
        - min_hotspot_width: Minimum width for a region to be called a hotspot.
    
        - signal_tr: Quantile threshold for outlier detection for background distribution fit.
        - adaptive_signal_tr: Use adaptive signal threshold. Signal_tr is ignored if True.

        - nonzero_window_fit: Minimum fraction of nonzero values in a background window to fit the model.
        - fdr_method: Method for FDR calculation. 'bh and 'by' are supported. 'bh' (default) is tested.
        - cpus: Number of CPUs to use. Won't use more than the number of chromosomes.

        - save_debug: Save debug information.
        - modwt_level: Level of MODWT decomposition. 7 is tested.
        - logger_level: Logging level.
    """
    window: int = 151
    min_mappable_bases_proportion: float = 0.05
    
    # Signal smoothing
    filter: str = 'haar'
    modwt_level: int = 7

    # Background model
    bg_window: int = 50001
    bg_window_small: int = 5001
    min_background_prop: float = 0.6
    max_background_prop: float = 0.995

    chromosome_fit_step: int = 1500

    # option to exclude peak flanks as well as peaks for fitting background
    # not used by default
    exclude_peak_flank_fit: int = 0

    # Remove peak flanks to reduce overfit to peak flanks when using local window
    exclude_peak_flank_scoring: int = 500

    # Remove obvious signal before segmentation
    signal_prop_sampling_step: int = 75
    # don't set less than signal_prop_sampling_step to avoid large memory usage
    signal_prop_interpolation_step: int = 1500 

    # Goodnes of fit calculation
    num_background_bins: int = 20
    num_signal_bins: int = 100
    min_obs_rmsea: int = 5 # Merge signal bins with less than this number of observations
    rmsea_tr: float = 0.05 # Threhold for good fit, used only to report 'bad' rmsea

    # Segmentation
    babachi_min_segment_size: int = 50001 # same as bg_window, can be changed
    babachi_segmentation_step: int = 500
    babachi_boundary_penalty: int = 20

    # Refit segments with signal/noise > outlier_segment_threshold * median(signal/noise)
    outlier_segment_threshold: float = 5
    max_outlier_iterations: int = 5

    # FDR correction
    fdr_method: str = 'bh'

    # Peak calling
    min_signif_bases_for_peak: int = 10
    min_peak_half_width: int = 10

    # Bigwigs params
    density_track_step: int = 20
    bg_track_step: int = 100

    # Utils
    cpus: int = 1
    tmp_dir: str = None
    save_debug: bool = False
    logger_level: int = logging.INFO


import psutil
import time
import subprocess
import sys
import os
import threading
import logging

from hotspot3.cli import parse_arguments
from hotspot3.io.logging import setup_logger

"""
This script is used to call main.py with memory tracking.
Inefficient to use this script in a pipeline, didn't optimize it too much.
Also, time estimates from logger might be inaccurate due to output buffering.
"""


def format_memory(size_in_bytes):
    """Format the memory size from bytes to a human-readable format."""
    for unit in ['B', 'KB', 'MB', 'GB', 'TB']:
        if size_in_bytes < 1024:
            return f"{size_in_bytes:.2f} {unit}"
        size_in_bytes /= 1024
    return f"{size_in_bytes:.2f} PB"


def track_memory(process, logger: logging.Logger, interval=2):
    """Track memory usage of a subprocess and log it."""
    python_process = psutil.Process(process.pid)
    logger.info("total_memory_rss_bytes\ttotal_memory_rss_human")
    
    while process.poll() is None:  # While the process is still running
    
        total_rss = python_process.memory_info().rss
        for child in python_process.children(recursive=True):
            if not child.is_running():
                continue
            total_rss += child.memory_info().rss

        total_rss_human = format_memory(total_rss)
        logger.info(f"{total_rss}\t{total_rss_human}")
        time.sleep(interval)


def run_process_with_memory_tracking(cmd, logger: logging.Logger):
    """Run a subprocess and track its memory usage."""
    try:
        process = subprocess.Popen(cmd)  # Start the process

        memory_thread = threading.Thread(target=track_memory, args=(process, logger))
        memory_thread.start()
        process.wait()
        memory_thread.join()
    except (psutil.NoSuchProcess, subprocess.SubprocessError, KeyboardInterrupt):
        logger.critical("Process interrupted or failed. Terminating...")
        raise
    finally:
        if memory_thread.is_alive():
            memory_thread.join()
        if process.poll() is None:
            process.terminate()
            try:
                process.wait(timeout=5)
            except subprocess.TimeoutExpired:
                process.kill()
                raise
        

def main():
    args, logger_level = parse_arguments(" with memory tracking. Creates {args.id}.memory_usage.tsv in output folder.")
    script_dir = os.path.dirname(os.path.abspath(__file__))
    cmd = ["python3", f"{script_dir}/main.py", *sys.argv[1:]]

    memory_log = f"{args.outdir}/{args.id}.memory_usage.tsv"
    logger = setup_logger(level=logger_level, outstream=memory_log)
    run_process_with_memory_tracking(cmd, logger)


if __name__ == "__main__":
    main()



import numpy as np
from scipy.signal import find_peaks


def find_stretches(arr: np.ndarray):
    """
    Find stretches of True values in a boolean array.

    Parameters:
        - arr: np.ndarray 1D boolean array.

    Returns:
        - start: np.ndarray of start positions of stretches
        - end: end positions of stretches
    """
    boundaries = np.diff(arr.astype(np.int8), prepend=0, append=0)
    start = np.where(boundaries == 1)[0]
    end = np.where(boundaries == -1)[0]
    return start, end


def filter_peaks_summits_within_regions(peaks_coordinates: np.ndarray, starts, ends):
    """
    Filter peaks which summits are within the regions defined by starts and ends.

    Parameters:
        - peaks_coordinates: np.array of shape (n_peaks, 3) with peak starts, summits and ends
        - starts: 1D array of regions' starts
        - ends: 1D array of regions' ends

    Returns:
        - filtered_peaks_mask: 1D boolean array of peaks that are within the regions
    """
    starts = np.asarray(starts)
    ends = np.asarray(ends)
    summits = peaks_coordinates[:, 1]

    closest_left_index = np.searchsorted(starts, summits, side='right') - 1
    filtered_peaks_mask = (closest_left_index >= 0) & (summits < ends[closest_left_index])
    
    return filtered_peaks_mask


def find_closest_min_peaks(signal):
    """
    Find peak summits (maxima) and peak starts and ends (closest local minima).
    Returns:
        - peaks_coordinates: np.array of shape (n_peaks, 3) with peak starts, summits and ends
    """
    maxima, _ = find_peaks(signal)
    minima, _ = find_peaks(-signal)
    total_dif = len(maxima) - (len(minima) - 1)
    if total_dif == 1:
        fist_minima_pos = minima[0]
        fist_maxima_pos = maxima[0]
        if fist_minima_pos < fist_maxima_pos:
            padding = (0, 1)
        else:
            padding = (1, 0)
    elif total_dif == 0:
        padding = 0
    else:
        padding = 1
    minima = np.pad(minima, padding, mode='constant', constant_values=(0, len(signal) - 1))
    peaks_coordinates = np.zeros([len(maxima), 3], dtype=int) # start, summit, end

    peaks_coordinates[:, 1] = maxima
    peaks_coordinates[:, 0] = minima[:len(maxima)]
    peaks_coordinates[:, 2] = minima[1:]
    return peaks_coordinates


def trim_at_threshold(signal, peaks_coordinates):
    """
    Trim peaks at the threshold height.
    Returns:
        - peaks_in_hotspots_trimmed: trimmed peaks coordinates
        - threshold_heights: heights of the threshold
    """
    heights = signal[peaks_coordinates]
    heights[:, 1] = heights[:, 1] / 2
    threshold_heights = np.max(heights, axis=1)

    peaks_in_hotspots_trimmed = np.zeros(peaks_coordinates.shape, dtype=np.int32)
    for i, (left, summit, right) in enumerate(peaks_coordinates):
        threshold = threshold_heights[i]
        new_left = summit - np.argmax(signal[left:summit + 1][::-1] <= threshold)

        new_right = summit + np.argmax(signal[summit:right + 1] <= threshold) - 1
        peaks_in_hotspots_trimmed[i, :] = np.array([new_left, summit, new_right])
    return peaks_in_hotspots_trimmed, threshold_heights


def find_varwidth_peaks(signal: np.ndarray, starts=None, ends=None):
    """
    Find variable width peaks within hotspots.

    Parameters:
        - signal: 1D array of smoothed signal values.
        - starts: 1D array of start positions of significant stretches.
        - ends: 1D array of end positions of significant stretches.
    
    Returns:
        - peaks_in_hotspots_trimmed: 2D array of start, summit, and end positions for each peak.
        - threshold_heights: 1D array of threshold heights for each peak.
    """
    if starts is None:
        assert ends is None
        starts = np.zeros(1)
        ends = np.full(1, len(signal))
    peaks_coordinates = find_closest_min_peaks(signal)
    in_sign_stretch = filter_peaks_summits_within_regions(peaks_coordinates, starts, ends)
    
    peaks_in_regions = peaks_coordinates[in_sign_stretch, :]

    peaks_in_hotspots_trimmed, threshold_heights = trim_at_threshold(signal, peaks_in_regions)
    
    return peaks_in_hotspots_trimmed, threshold_heights




"""
Low level classes to fit background.
GlobalBackgroundFit: Fit the whole array.
WindowBackgroundFit: Fit the array in a windowed fashion.
StridedBackgroundFit: Fit the array using a strided approach (core functionality is now defunc).
Currently, it is used only to calculate sliding quantile
"""

import numpy.ma as ma
import numpy as np
from scipy.special import betainc
from typing import List
import warnings

from hotspot3.helpers.models import NotEnoughDataForContig, FitResults, WindowedFitResults, DataForFit
from hotspot3.config import ProcessorConfig
from hotspot3.connectors.bottleneck import BottleneckWrapper
from hotspot3.background_fit import calc_rmsea, rolling_view_with_nan_padding
from hotspot3.helpers.stats import check_valid_nb_params
from hotspot3.helpers.utils import wrap_masked


class BackgroundFit(BottleneckWrapper):
    def __init__(self, config: ProcessorConfig=None, logger=None, name=None):
        super().__init__(logger=logger, config=config, name=name)

        self.sampling_step = self.config.signal_prop_sampling_step
        self.interpolation_step = self.config.signal_prop_interpolation_step // self.sampling_step

    def fit(self):
        raise NotImplementedError
    
    def prepare_data_for_fit(self, array: np.ndarray):
        raise NotImplementedError
    
    @wrap_masked
    def get_mean_and_var(self, array: np.ndarray, **kwargs):
        with np.errstate(invalid='ignore', divide='ignore'):
            mean = np.nanmean(array, axis=0, dtype=np.float32, **kwargs)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)
                var = np.nanvar(array, ddof=1, axis=0, dtype=np.float32, **kwargs)
        return mean, var

    @wrap_masked
    def p_from_mean_and_r(self, mean: np.ndarray, r: np.ndarray, where=None):
        with np.errstate(divide='ignore', invalid='ignore', all='ignore'):
            p = np.array(mean / (mean + r), dtype=np.float16) # check if we return p or 1-p
        return p
    
    @wrap_masked
    def r_from_mean_and_var(self, mean: np.ndarray, var: np.ndarray):
        with np.errstate(divide='ignore', invalid='ignore', all='ignore'):
            r = np.array(mean ** 2 / (var - mean), dtype=np.float16)
        return r
    
    def value_counts_per_bin(self, array, bin_edges, where=None):
        if where is None:
            where = np.ones(array.shape, dtype=bool)
        left_edge = bin_edges[0]
        right_edges = bin_edges[1:]
        value_counts = np.full_like(right_edges, 0, dtype=np.float32)
        for i in range(right_edges.shape[0]):
            right_edge = right_edges[i]
            bin_membership = (array >= left_edge[None, :]) & (array < right_edge[None, :])
            value_counts[i] = np.sum(bin_membership, axis=0, where=where)
            left_edge = right_edge
        
        return value_counts
    
    def get_all_bins(self, array: np.ndarray, fallback_fit_results: FitResults=None):
        min_bg_tr = self.quantile_ignore_all_na(array, self.config.min_background_prop)

        signal_bins, n_signal_bins = self.get_signal_bins(
            array,
            min_bg_tr=min_bg_tr,
            fallback_fit_results=fallback_fit_results
        )
        n_bg_bins = min(np.nanmax(min_bg_tr), self.config.num_background_bins)
        n_bg_bins = round(n_bg_bins)

        bg_bins = np.full((n_bg_bins + 1, *min_bg_tr.shape), np.nan)
        bg_bins[:, ~np.isnan(min_bg_tr)] = np.round(
            np.linspace(
                0,
                min_bg_tr[~np.isnan(min_bg_tr)],
                n_bg_bins + 1,
            )
        )

        
        return np.concatenate([bg_bins[:-1], signal_bins]), n_signal_bins

    def get_signal_bins(self, array: np.ndarray, min_bg_tr=None, fallback_fit_results: FitResults=None):
        if min_bg_tr is None:
            min_bg_tr = self.quantile_ignore_all_na(array, self.config.min_background_prop)

        if self.config.max_background_prop <= self.config.min_background_prop:
            max_bg_tr = min_bg_tr + 1
        else:
            max_bg_tr = self.quantile_ignore_all_na(array, self.config.max_background_prop)

        if fallback_fit_results is not None:
            max_bg_tr = np.maximum(
                max_bg_tr,
                fallback_fit_results.fit_threshold
            )
        n_signal_bins = min(np.nanmax(max_bg_tr - min_bg_tr), self.config.num_signal_bins)
        n_signal_bins = round(n_signal_bins)
        
        bin_edges = np.full((n_signal_bins + 1, *min_bg_tr.shape), np.nan, dtype=np.float32)
        nan_tr = np.isnan(min_bg_tr) | np.isnan(max_bg_tr)
        with np.errstate(invalid='ignore'):
            bin_edges[:, ~nan_tr] = np.round(
                np.linspace(min_bg_tr[~nan_tr], max_bg_tr[~nan_tr], n_signal_bins + 1)
            )
            bin_edges[0, ~nan_tr] = min_bg_tr[~nan_tr]

        return bin_edges, n_signal_bins
    
    def merge_signal_bins_for_rmsea_inplace(self, value_counts, bin_edges, n_signal_bins):
        assert bin_edges.shape[0] == value_counts.shape[0] + 1, f"Bin edges shape should be one more than value counts shape. Got {bin_edges.shape} and {value_counts.shape}"
        for j in np.arange(n_signal_bins):
            i = value_counts.shape[0] - j - 1

            condition = value_counts[i] < self.config.min_obs_rmsea
            value_counts[i - 1] += np.where(
                condition,
                value_counts[i],
                0
            )
            bin_edges[i] = np.where(
                condition,
                bin_edges[i + 1],
                bin_edges[i]
            )

            value_counts[i] = np.where(
                condition,
                0,
                value_counts[i]
            )

    def quantile_ignore_all_na(self, array: np.ndarray, quantile: float):
        """
        Wrapper of nan quantile to avoid runtime warnings when data is all nan.
        """
        all_nan = np.all(np.isnan(array), axis=0)
        if array.ndim == 1:
            return np.nan if all_nan else np.nanquantile(array, quantile)
        result = np.full(array.shape[1], np.nan, dtype=np.float32)
        result[~all_nan] = np.nanquantile(array[:, ~all_nan], quantile, axis=0, method='higher')
        return result
    
    @wrap_masked
    def calc_rmsea_all_windows(
        self,
        p: np.ndarray,
        r: np.ndarray,
        n_params: int,
        bin_edges: np.ndarray, # edges of the bins, right edge not inclusive
        value_counts_per_bin: np.ndarray, # number of observed cutcounts in each bin for each window
    ):
        """
        Calculate RMSEA for all sliding windows.
        """
        valid_bins = np.ones_like(value_counts_per_bin, dtype=bool)
        bg_sum_mappable = np.sum(value_counts_per_bin, axis=0, where=valid_bins)

        sf_values = np.where(bin_edges == 0, 1., betainc(bin_edges, r, p))
        sf_diffs = -np.diff(sf_values, axis=0)
       
        assert sf_diffs.shape == value_counts_per_bin.shape, f"SF diffs shape should match value counts shape. Got SF: {sf_diffs.shape} and vc: {value_counts_per_bin.shape}"
        norm_coef = 1 - sf_values[-1]
        expected_counts = (sf_diffs * bg_sum_mappable / norm_coef)
        with np.errstate(invalid='ignore'):
            df = np.sum(
                np.diff(bin_edges, axis=0) != 0,
                axis=0,
                where=valid_bins
            ) - n_params - 1
        rmsea = calc_rmsea(
            value_counts_per_bin,
            expected_counts,
            bg_sum_mappable,
            df,
            where=valid_bins
        )
        return rmsea

    def get_bg_quantile_from_tr(self, agg_cutcounts: np.ndarray, tr: float):
        with np.errstate(invalid='ignore'):
            result = np.sum(agg_cutcounts < tr, axis=0) / np.sum(~np.isnan(agg_cutcounts), axis=0)
        return result


class GlobalBackgroundFit(BackgroundFit):
    """
    Class to fit the background distribution globally (for chromosome/homogeneous regions)
    """
    def fit(
            self,
            agg_cutcounts: ma.MaskedArray,
            step=None,
            fallback_fit_results: FitResults = None
        ) -> FitResults:
        """
        Fit the global background distribution.

        Parameters:
            agg_cutcounts (np.ndarray): Array of aggregated cutcounts.
            step (int): Step to reduce computational burden and improve speed. Can be set to 1 for full resolution.
            fallback_fit_results (FitResults): Fallback fit results to use if the fit has failed.
        """

        data_for_fit = self.prepare_data_for_fit(agg_cutcounts, step, fallback_fit_results)
        best_fit_result = self.fit_and_choose_best(data_for_fit)
        if not check_valid_nb_params(best_fit_result):
            if fallback_fit_results is not None:
                best_fit_result = self.fit_and_choose_best(data_for_fit, fallback_fit_results)
            else:
                raise NotEnoughDataForContig
        best_fit_result.n_total = agg_cutcounts.count()
        signal_mask = self.get_signal_mask_for_tr(
            agg_cutcounts,
            best_fit_result.fit_threshold
        )
        best_fit_result.n_signal = signal_mask.sum()

        best_fit_result.signal_tags = agg_cutcounts[signal_mask].compressed().astype(np.int64).sum()
       
        best_fit_result.total_tags = np.nansum(agg_cutcounts[np.isfinite(agg_cutcounts)].compressed().astype(np.int64))

        return best_fit_result
    
    def fit_and_choose_best(self, data_for_fit: DataForFit, fallback_fit_results: FitResults=None):
        result = self.fit_all_thresholds(data_for_fit, fallback_fit_results)
        if len(result) == 0: # No valid fits => fit all data
            best_fit_result = self.fit_for_tr(
                data_for_fit,
                np.inf,
                fallback_fit_results=fallback_fit_results,
                calc_rmsea=False
            )
        else:
            best_fit_result = min(result, key=lambda x: x.rmsea)
        return best_fit_result
    
    def prepare_data_for_fit(
            self,
            agg_cutcounts: ma.MaskedArray,
            step: int=None,
            fallback_fit_results: FitResults=None
        ):
        if step is None:
            step = self.config.window
        agg_cutcounts = agg_cutcounts.filled(np.nan)
        max_counts_with_flanks = self.get_max_count_with_flanks(agg_cutcounts)[::step]
        agg_cutcounts = agg_cutcounts[::step]
        bin_edges, n_signal_bins = self.get_all_bins(agg_cutcounts, fallback_fit_results)
        bin_edges = bin_edges[:, None]
        value_counts = self.value_counts_per_bin(agg_cutcounts[:, None], bin_edges)
        self.merge_signal_bins_for_rmsea_inplace(value_counts, bin_edges, n_signal_bins)
        return DataForFit(
            bin_edges,
            value_counts,
            n_signal_bins,
            agg_cutcounts,
            max_counts_with_flanks
        )
    
    def fit_all_thresholds(
            self,
            data_for_fit: DataForFit,
            fallback_fit_results: FitResults=None
            ) -> List[FitResults]:
        result = []
        for i in np.arange(data_for_fit.n_signal_bins)[::-1]:
            tr = data_for_fit.bin_edges[len(data_for_fit.bin_edges) - i - 1, 0]
            try:
                assumed_signal_mask = data_for_fit.max_counts_with_flanks >= tr
                step_fit = self._fit_for_bg_mask(
                    data_for_fit.agg_cutcounts,
                    bin_edges=data_for_fit.bin_edges,
                    assumed_signal_mask=assumed_signal_mask,
                    fallback_fit_results=fallback_fit_results
                )
            except NotEnoughDataForContig:
                continue
            q = self.get_bg_quantile_from_tr(data_for_fit.agg_cutcounts, tr)
            step_fit.fit_quantile = q
            step_fit.fit_threshold = tr
            result.append(step_fit)
        return result
    
    @wrap_masked
    def fit_for_tr(
            self,
            data_for_fit: DataForFit,
            tr: float,
            fallback_fit_results: FitResults=None,
            calc_rmsea=True
    ):
        assumed_signal_mask = data_for_fit.max_counts_with_flanks >= tr

        step_fit = self._fit_for_bg_mask(
            data_for_fit.agg_cutcounts,
            assumed_signal_mask,
            data_for_fit.bin_edges,
            fallback_fit_results=fallback_fit_results,
            calc_rmsea=calc_rmsea
        )
        step_fit.fit_quantile = self.get_bg_quantile_from_tr(data_for_fit.agg_cutcounts, tr)
        step_fit.fit_threshold = tr
        return step_fit
    
    def _fit_for_bg_mask(
        self,
        agg_cutcounts,
        assumed_signal_mask,
        bin_edges,
        fallback_fit_results: FitResults=None,
        calc_rmsea=True
    ):
        mean, var = self.get_mean_and_var(agg_cutcounts, where=~assumed_signal_mask)
        if fallback_fit_results is not None:
            r = fallback_fit_results.r
            n_params = 1
        else:
            r = self.r_from_mean_and_var(mean, var)
            n_params = 2
        p = self.p_from_mean_and_r(mean, r)

        if not check_valid_nb_params(FitResults(p, r)):
            raise NotEnoughDataForContig
        
        if calc_rmsea: # Used to bypass RMSEA calculation
            # in the code used when np.inf passed 
            value_counts = self.value_counts_per_bin(
                agg_cutcounts[:, None],
                bin_edges,
                where=~assumed_signal_mask[:, None]
            )
            rmsea = self.calc_rmsea_all_windows(
                p,
                r,
                n_params,
                bin_edges,
                value_counts,
            )[0]
            if not np.isfinite(rmsea):
                raise NotEnoughDataForContig
        else:
            rmsea = np.nan
        return FitResults(p, r, rmsea)


class StridedBackgroundFit(BackgroundFit):
    """
    Class to fit the background distribution using a strided approach.
    Extemely computationally taxing, use large stride values to compensate.
    """

    def find_thresholds_at_chrom_quantile(
            self,
            agg_cutcounts: np.ndarray,
            fit_quantile: float,
            bg_window=None
        ):
        original_shape = agg_cutcounts.shape
        strided_agg_cutcounts = self.get_strided_agg_cutcounts(agg_cutcounts, bg_window=bg_window)
        subsampled_indices = self.get_subsampled_indices(original_shape)
        values = self.quantile_ignore_all_na(
            strided_agg_cutcounts,
            fit_quantile
        )
        return self.upcast(original_shape, subsampled_indices, values)

    def get_strided_agg_cutcounts(self, agg_cutcounts: np.ndarray, bg_window=None):
        if bg_window is None:
            bg_window = self.config.bg_window
        points_in_bg_window = (bg_window - 1) // self.sampling_step
        if points_in_bg_window % 2 == 0:
            points_in_bg_window += 1
        
        strided_agg_cutcounts = rolling_view_with_nan_padding(
            agg_cutcounts[::self.sampling_step],
            points_in_window=points_in_bg_window,
            interpolation_step=self.interpolation_step
        ) # shape (bg_window, n_points)

        return strided_agg_cutcounts
    
    def get_subsampled_indices(self, original_shape: tuple):
        return np.arange(0, original_shape[0], self.sampling_step, dtype=np.uint32)[::self.interpolation_step]
    
    def upcast(self, original_shape, subsampled_indices: np.ndarray, values: np.ndarray):
        upcasted = np.full(original_shape, np.nan, dtype=np.float16)
        with np.errstate(over='ignore'):
            upcasted[subsampled_indices] = np.asarray(values, dtype=np.float16)
        upcasted[~np.isfinite(upcasted)] = np.nan
        return upcasted


class WindowBackgroundFit(BackgroundFit):
    """
    Class to fit the background distribution in a running window fashion
    """
    def fit(self, array: ma.MaskedArray, per_window_trs, fallback_fit_results: FitResults=None) -> WindowedFitResults:
        agg_cutcounts = array.copy()

        high_signal_mask = self.get_signal_mask_for_tr(
            agg_cutcounts,
            per_window_trs,
            flank_length=self.config.exclude_peak_flank_scoring
        )
        agg_cutcounts[high_signal_mask] = np.nan
        fit_results = self.sliding_method_of_moments_fit(
            agg_cutcounts,
            fallback_fit_results=fallback_fit_results
        )

        return fit_results
    
    def sliding_method_of_moments_fit(
            self,
            agg_cutcounts: ma.MaskedArray,
            fallback_fit_results: FitResults
        ):

        bg_window = self.config.bg_window
        mean = self.centered_running_nanmean(agg_cutcounts, bg_window)
        
        mean = ma.masked_invalid(mean)
        enough_bg_mask = ~mean.mask
        
        if fallback_fit_results is not None:
            r = np.full_like(mean, fallback_fit_results.r, dtype=np.float32)
        else:
            var = self.centered_running_nanvar(agg_cutcounts, bg_window)
            var = ma.masked_invalid(var)
            r = self.r_from_mean_and_var(mean, var).filled(np.nan)

        p = self.p_from_mean_and_r(mean, r).filled(np.nan)
        return WindowedFitResults(p, r, enough_bg_mask=enough_bg_mask)
